from operator import ge, gt, le, lt


class Comparable:
    """
    Base class that allows subclasses to perform arbitrary comparison operations by implementing the necessary interfaces.

    Subclasses need to implement the `_key` method, which returns the key used for comparison.

    Supported comparison operators: ==, !=, >=, >, <=, <

    Usage:
        * Subclass `Comparable` and implement the `_key` method.
        * Objects of the subclass can then be compared using the defined comparison operators.

    Example:
        >>> class MyComparable(Comparable):
        >>>     def __init__(self, value):
        >>>         self.value = value
        ...
        >>>     def _key(self):
        >>>         return self.value
        ...
        >>> obj1 = MyComparable(5)
        >>> obj2 = MyComparable(10)
        >>> print(obj1 == obj2)  # False
        >>> print(obj1 < obj2)   # True

    """

    def _check_same_type(self, other):
        """
        Helper method to check if the type of the other object is the same as the current object's type.

        :param other: The other object to compare.
        :type other: object
        :returns: The other object if the types are the same.
        :rtype: object
        :raises TypeError: If the types are different.
        """
        if type(self) == type(other):
            return other
        else:
            raise TypeError(f'Invalid type for compare - {other!r}.')

    def _key(self):
        """
        Abstract method that needs to be implemented by subclasses.
        Returns the key used for comparison.

        :returns: The key used for comparison.
        :rtype: object
        :raises NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError  # pragma: no cover

    def __eq__(self, other):
        """
        Implementation of the equality operator (==).

        :param other: The other object to compare.
        :type other: object
        :returns: True if the objects are equal, False otherwise.
        :rtype: bool
        """
        try:
            other = self._check_same_type(other)
        except TypeError:
            return False
        else:
            return self._key() == other._key()

    def __ne__(self, other):
        """
        Implementation of the inequality operator (!=).

        :param other: The other object to compare.
        :type other: object
        :returns: True if the objects are not equal, False otherwise.
        :rtype: bool
        """
        try:
            other = self._check_same_type(other)
        except TypeError:
            return True
        else:
            return self._key() != other._key()

    def _cmp(self, cmp, other):
        """
        Helper method to perform comparison operations using the provided comparison function.

        :param cmp: The comparison function to use.
        :type cmp: callable
        :param other: The other object to compare.
        :type other: object
        :returns: The result of the comparison operation.
        :rtype: bool
        """
        try:
            other = self._check_same_type(other)
        except TypeError:
            return False
        else:
            return cmp(self._key(), other._key())

    def __ge__(self, other):
        """
        Implementation of the greater than or equal to operator (>=).

        :param other: The other object to compare.
        :type other: object
        :returns: True if the current object is greater than or equal to the other object, False otherwise.
        :rtype: bool
        """
        return self._cmp(ge, other)

    def __gt__(self, other):
        """
        Implementation of the greater than operator (>).

        :param other: The other object to compare.
        :type other: object
        :returns: True if the current object is greater than the other object, False otherwise.
        :rtype: bool
        """
        return self._cmp(gt, other)

    def __le__(self, other):
        """
        Implementation of the less than or equal to operator (<=).

        :param other: The other object to compare.
        :type other: object
        :returns: True if the current object is less than or equal to the other object, False otherwise.
        :rtype: bool
        """
        return self._cmp(le, other)

    def __lt__(self, other):
        """
        Implementation of the less than operator (<).

        :param other: The other object to compare.
        :type other: object
        :returns: True if the current object is less than the other object, False otherwise.
        :rtype: bool
        """
        return self._cmp(lt, other)



import random
import time
from typing import Optional, Dict

import requests
from requests.adapters import HTTPAdapter, Retry
from requests.exceptions import RequestException

DEFAULT_TIMEOUT = 10  # seconds


class TimeoutHTTPAdapter(HTTPAdapter):
    """
    Custom HTTP adapter that sets a default timeout for requests.

    Inherits from `HTTPAdapter`.

    Usage:
    - Create an instance of `TimeoutHTTPAdapter` and pass it to a `requests.Session` object's `mount` method.

    Example:
    ```python
    session = requests.Session()
    adapter = TimeoutHTTPAdapter(timeout=10)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    ```

    :param timeout: The default timeout value in seconds. (default: 10)
    :type timeout: int
    """

    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        """
        Sends a request with the provided timeout value.

        :param request: The request to send.
        :type request: PreparedRequest
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :returns: The response from the request.
        :rtype: Response
        """
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def get_requests_session(max_retries: int = 5, timeout: int = DEFAULT_TIMEOUT,
                         headers: Optional[Dict[str, str]] = None, session: Optional[requests.Session] = None) \
        -> requests.Session:
    """
    Returns a requests Session object configured with retry and timeout settings.

    :param max_retries: The maximum number of retries. (default: 5)
    :type max_retries: int
    :param timeout: The default timeout value in seconds. (default: 10)
    :type timeout: int
    :param headers: Additional headers to be added to the session. (default: None)
    :type headers: Optional[Dict[str, str]]
    :param session: An existing requests Session object to use. If not provided, a new Session object is created. (default: None)
    :type session: Optional[requests.Session]
    :returns: The requests Session object.
    :rtype: requests.Session
    """
    session = session or requests.session()
    retries = Retry(
        total=max_retries, backoff_factor=1,
        status_forcelist=[413, 429, 500, 501, 502, 503, 504, 505, 506, 507, 509, 510, 511],
        allowed_methods=["HEAD", "GET", "POST", "PUT", "DELETE", "OPTIONS", "TRACE"],
    )
    adapter = TimeoutHTTPAdapter(max_retries=retries, timeout=timeout)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    session.headers.update({
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
                      "(KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        **dict(headers or {}),
    })

    return session


def srequest(session: requests.Session, method, url, *, max_retries: int = 5,
             sleep_time: float = 5.0, raise_for_status: bool = True, **kwargs) -> requests.Response:
    """
    Send a request using the provided session object with retry and timeout settings.

    :param session: The requests Session object to use for the request.
    :type session: requests.Session
    :param method: The HTTP method for the request.
    :type method: str
    :param url: The URL for the request.
    :type url: str
    :param max_retries: The maximum number of retries. (default: 5)
    :type max_retries: int
    :param sleep_time: The sleep time between retries in seconds. (default: 5.0)
    :type sleep_time: float
    :param raise_for_status: Whether to raise an exception for non-successful response status codes. (default: True)
    :type raise_for_status: bool
    :param kwargs: Additional keyword arguments for the request.
    :type kwargs: dict
    :returns: The response from the request.
    :rtype: requests.Response
    """
    if isinstance(session, (list, tuple)):
        session = random.choice(session)

    resp = None
    for _ in range(max_retries):
        try:
            resp = session.request(method, url, **kwargs)
        except RequestException:
            time.sleep(sleep_time)
        else:
            break
    assert resp is not None, f'Request failed for {max_retries} time(s).'
    if raise_for_status:
        resp.raise_for_status()

    return resp


def sget(session: requests.Session, url, *, max_retries: int = 5,
         sleep_time: float = 5.0, raise_for_status: bool = True, **kwargs) -> requests.Response:
    """
    Send a GET request using the provided session object with retry and timeout settings.

    :param session: The requests Session object to use for the request.
    :type session: requests.Session
    :param url: The URL for the request.
    :type url: str
    :param max_retries: The maximum number of retries. (default: 5)
    :type max_retries: int
    :param sleep_time: The sleep time between retries in seconds. (default: 5.0)
    :type sleep_time: float
    :param raise_for_status: Whether to raise an exception for non-successful response status codes. (default: True)
    :type raise_for_status: bool
    :param kwargs: Additional keyword arguments for the request.
    :type kwargs: dict
    :returns: The response from the request.
    :rtype: requests.Response
    """
    return srequest(
        session, 'GET', url,
        max_retries=max_retries,
        sleep_time=sleep_time,
        raise_for_status=raise_for_status,
        **kwargs,
    )



from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName


class ChineseName(_GenericChineseName):
    pass


class EnglishName(_GenericEnglishName):
    pass



from typing import List

SITES = {
    'anime_pictures': 'anime-pictures.net',
    'atfbooru': 'booru.allthefallen.moe',
    'sankaku': 'chan.sankakucomplex.com',
    'danbooru': 'danbooru.donmai.us',
    'hypnohub': 'hypnohub.net',
    'konachan': 'konachan.com',
    'konachan_net': 'konachan.net',
    'lolibooru': 'lolibooru.moe',
    'rule34': 'rule34.xxx',
    'safebooru': 'safebooru.donmai.us',
    'xbooru': 'xbooru.com',
    'yande': 'yande.re',
    'zerochan': 'zerochan.net',
}


def list_available_sites() -> List[str]:
    """
    List the available supported image website names.

    :return: The list of supported website names.
    :rtype: List[str]

    Examples::
        >>> from gchar.resources.sites import list_available_sites
        >>>
        >>> list_available_sites()
        ['anime_pictures', 'atfbooru', 'danbooru', 'hypnohub', 'konachan', 'konachan_net', 'lolibooru', 'rule34', 'safebooru', 'sankaku', 'xbooru', 'yande', 'zerochan']
    """
    return sorted(SITES.keys())



from enum import Enum


class Rarity(Enum):
    R = ('普', 1)
    SR = ('危', 2)
    SSR = ('狂', 3)

    def __init__(self, text: str, number: int):
        self.number = number
        self.text = text

    @classmethod
    def loads(cls, obj) -> 'Rarity':
        if isinstance(obj, Rarity):
            return obj
        elif isinstance(obj, int):
            for name, item in cls.__members__.items():
                if item.number == obj:
                    return item
            raise ValueError(f'Unknown rarity number - {obj!r}.')
        elif isinstance(obj, str):
            for name, item in cls.__members__.items():
                if name.upper() == obj.upper() or item.text.upper() == obj.upper():
                    return item
            raise KeyError(f'Unknown rarity text - {obj!r}.')
        else:
            raise TypeError(f'Unknown rarity type - {obj!r}.')

    def __eq__(self, other):
        if isinstance(other, Rarity):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError, KeyError):
                return False



import math
from typing import List, Tuple

from .name import EnglishName, JapaneseName, ChineseName, ChineseAliasName
from .property import Rarity, Clazz
from ..base import Character as _BaseCharacter

_KNOWN_DATA_FIELDS = [
    'data-adapt', 'data-atk', 'data-birth_place', 'data-block', 'data-cost', 'data-def', 'data-en',
    'data-flex', 'data-group', 'data-hp', 'data-id', 'data-interval', 'data-ja', 'data-logo',
    'data-nation', 'data-obtain_method', 'data-phy', 'data-plan', 'data-position', 'data-potential',
    'data-profession', 'data-race', 'data-rarity', 'data-re_deploy', 'data-res', 'data-sex',
    'data-skill', 'data-sortid', 'data-subprofession', 'data-tag', 'data-team', 'data-tolerance',
    'data-trust', 'data-zh'
]


class Character(_BaseCharacter):
    """
    A class for modeling characters in the Arknights game.

    Here is an example of operator data from prts.wiki

    .. code-block:: json
        :linenos:

        {
            "data-adapt": "普通",
            "data-atk": "102",
            "data-birth_place": "未公开",
            "data-block": "1",
            "data-cost": "3",
            "data-def": "28",
            "data-en": "U-Official",
            "data-flex": "缺陷",
            "data-group": "",
            "data-hp": "385",
            "data-id": "U007",
            "data-interval": "1.3s",
            "data-ja": "",
            "data-logo": "罗德岛",
            "data-nation": "罗德岛",
            "data-obtain_method": "活动获得",
            "data-phy": "普通",
            "data-plan": "缺陷",
            "data-position": "远程位",
            "data-potential": "`",
            "data-profession": "辅助",
            "data-race": "札拉克",
            "data-rarity": "0",
            "data-re_deploy": "200s",
            "data-res": "0",
            "data-sex": "女",
            "data-skill": "缺陷",
            "data-sortid": "274",
            "data-subprofession": "吟游者",
            "data-tag": "控场",
            "data-team": "",
            "data-tolerance": "普通",
            "data-trust": "120,20,0",
            "data-zh": "U-Official"
        }
    """
    __game_name__ = 'arknights'
    __official_name__ = 'Arknights'
    __cnname_class__ = ChineseName
    __enname_class__ = EnglishName
    __jpname_class__ = JapaneseName
    __alias_name_class__ = ChineseAliasName

    def __init__(self, raw_data: dict):
        """
        Initialize a character with the raw data.

        :param raw_data: The raw data of the character.
        :type raw_data: dict
        """
        self.__origin_raw_data = raw_data
        self.__raw_data = raw_data['data']
        self.__skins = raw_data['skins']
        self.__is_extra = None

    def _index(self):
        return self.__raw_data.get('data-index') or self.__raw_data.get('data-id')

    def _gender(self):
        return self.__raw_data['data-sex']

    @property
    def rarity(self) -> Rarity:
        """
        Get the rarity of the character.

        :return: The rarity of the character.
        :rtype: :class:`Rarity`
        """
        return Rarity.loads(int(self.__raw_data['data-rarity']) + 1)

    @property
    def clazz(self) -> Clazz:
        """
        Get the class of the character.

        :return: The class of the character.
        :rtype: Clazz
        """
        return Clazz.loads(self.__raw_data.get('data-class') or self.__raw_data.get('data-profession'))

    def _cnname(self):
        return self.__raw_data.get('data-cn') or self.__raw_data.get('data-zh') or None

    def _enname(self):
        return self.__raw_data.get('data-en') or None

    def _jpname(self):
        return self.__raw_data.get('data-jp') or self.__raw_data.get('data-ja') or None

    def _custom_alias_names(self):
        return list(self.__origin_raw_data.get('alias', []) or [])

    def _skins(self) -> List[Tuple[str, str]]:
        return [(item['name'], item['url']) for item in self.__skins]

    def __getattr__(self, item: str):
        key = 'data-' + item.replace('_', '-')
        if key in _KNOWN_DATA_FIELDS:
            return self.__raw_data.get(key, None)
        else:
            return object.__getattribute__(self, item)

    def _is_extra(self) -> bool:
        return (self.enname and 'the' in self.enname) or \
            (self.enname == 'amiya' and self.cnname != '阿米娅')

    def _order(self):
        release_info = self.__origin_raw_data['release']
        _release_time = release_info['time'] if release_info['time'] is not None else math.inf
        _release_index = release_info['index'] if release_info['index'] is not None else math.inf
        return _release_time, _release_index

    def _release_time(self):
        _release_time, _ = self._order()
        return _release_time

    def __repr__(self):
        return f'<{type(self).__name__} {self.index} - {"/".join(map(str, self._names()))}, ' \
               f'{self.gender.name.lower()}, {self.rarity}{"*" * self.rarity}>'



_GAMES = {
    'arknights': ('アークナイツ', 'アークナイツ'),
    'fgo': ('Fate/GrandOrder', 'Fate'),
    'azurlane': ('アズールレーン', 'アズールレーン'),
    'genshin': ('原神', '原神'),
    'girlsfrontline': ('ドールズフロントライン', 'ドールズフロントライン'),
    'neuralcloud': ('云图计划', '云图计划'),
    'bluearchive': ('ブルーアーカイブ', 'ブルーアーカイブ'),
    'pathtonowhere': ('無期迷途', '無期迷途'),
    'nikke': ('勝利の女神:NIKKE', '勝利の女神:NIKKE'),
    'starrail': ('崩壊スターレイル', 'スターレイル'),
}



from enum import IntEnum


class Rarity(IntEnum):
    SR = 4
    SSR = 5

    @classmethod
    def loads(cls, obj):
        if isinstance(obj, Rarity):
            return obj
        elif isinstance(obj, str):
            return cls.__members__[obj.upper().strip()]
        elif isinstance(obj, int):
            for name, item in cls.__members__.items():
                if item.value == obj:
                    return item

            raise ValueError(f'Unknown rarity value - {obj!r}.')
        else:
            raise TypeError(f'Unknown rarity type - {obj!r}.')



import re
from typing import List, Union

from ...utils import Comparable


class _BaseName(Comparable):
    """
    Base class for modeling character names.

    The class provides common functionality for character name modeling.
    It inherits from the `Comparable` class and defines methods for equality comparison and string representation.

    Usage:
    - Use the `_BaseName` class as a base class for specific character name modeling classes.

    :param name: The name of the character.
    :type name: str
    """

    def __init__(self, name):
        """
        Initialize the _BaseName object.

        :param name: The name of the character.
        :type name: str
        """
        _ = name

    def _repr(self) -> str:
        """
        Get the string representation of the character name.

        This method should be implemented by the subclasses.

        :returns: The string representation of the character name.
        :rtype: str
        """
        raise NotImplementedError  # pragma: no cover

    def _check_same_type(self, other):
        """
        Check if the other object is of the same type as the current object.

        If the other object is of the same type, it is returned as is.
        Otherwise, a new instance of the current class is created with the value of the other object.

        :param other: The other object to check.
        :type other: object
        :returns: The other object if it is of the same type, or a new instance of the current class with the value of the other object.
        :rtype: _BaseName
        """
        if isinstance(other, _BaseName):
            return other
        else:
            return self.__class__(other)

    def _key(self):
        """
        Get the key used for comparison.

        This method should be implemented by the subclasses.

        :returns: The key used for comparison.
        :rtype: object
        """
        return self._repr()

    def __str__(self):
        """
        Get the string representation of the character name.

        :returns: The string representation of the character name.
        :rtype: str
        """
        return self._repr()

    def __repr__(self):
        """
        Get the string representation of the _BaseName object.

        :returns: The string representation of the _BaseName object.
        :rtype: str
        """
        return f'<{type(self).__name__} {self._repr()!r}>'

    def __bool__(self):
        """
        Check if the character name is considered True.

        :returns: True if the character name is considered True, False otherwise.
        :rtype: bool
        """
        return bool(self._repr())


class TextName(_BaseName):
    """
    Class for modeling character names based on text.

    The class inherits from _BaseName and provides methods for string representation and comparison.
    It supports containment check, length calculation, and item retrieval.

    Usage:
        * Use the TextName class to model character names based on text.

    :param name: The name of the character.
    :type name: str
    """

    def __init__(self, name: str):
        """
        Initialize the TextName object.

        :param name: The name of the character.
        :type name: str
        """
        if not isinstance(name, str):
            raise TypeError(f'Invalid name type - {name!r}.')
        _BaseName.__init__(self, name)
        self.__name = self._preprocess(name)

    def _repr(self) -> str:
        """
        Get the string representation of the character name.

        :returns: The string representation of the character name.
        :rtype: str
        """
        return self.__name

    def __contains__(self, item: str) -> bool:
        """
        Check if the character name contains a specific string.

        :param item: The string to check.
        :type item: str
        :returns: True if the character name contains the specified string, False otherwise.
        :rtype: bool
        """
        return item in self.__name

    def __len__(self):
        """
        Calculate the length of the character name.

        :returns: The length of the character name.
        :rtype: int
        """
        return len(self.__name)

    def __getitem__(self, item):
        """
        Get the item at the specified index or slice.

        :param item: The index or slice to retrieve.
        :type item: Union[int, slice]
        :returns: The item at the specified index or slice.
        :rtype: Any
        """
        return self.__name[item]

    @classmethod
    def _preprocess(cls, name: str) -> str:
        """
        Preprocess the character name.

        This method should be implemented by the subclasses to provide custom preprocessing logic.

        :param name: The name of the character.
        :type name: str
        :returns: The preprocessed name.
        :rtype: str
        """
        return name.strip()

    @classmethod
    def _eqprocess(cls, name: str) -> str:
        """
        Preprocess the character name for equality comparison.

        This method should be implemented by the subclasses to provide custom preprocessing logic.

        :param name: The name of the character.
        :type name: str
        :returns: The preprocessed name for equality comparison.
        :rtype: str
        """
        return re.sub(r'[\W_]+', '', name.lower())

    def _key(self):
        """
        Get the key used for comparison.

        :returns: The key used for comparison.
        :rtype: object
        """
        return self._eqprocess(_BaseName._key(self))


class SegmentName(_BaseName):
    """
    Class for modeling character names based on segments.

    The class inherits from _BaseName and provides methods for string representation and comparison.
    It supports containment check, length calculation, and item retrieval.

    Usage:
        * Use the SegmentName class to model character names based on segments.

    :param name: The name of the character.
    :type name: Union[str, List[str]]
    """
    __seperator__ = '_'

    def __init__(self, name: Union[str, List[str]]):
        """
        Initialize the SegmentName object.

        :param name: The name of the character.
        :type name: Union[str, List[str]]
        """
        if not isinstance(name, (str, list, tuple)):
            raise TypeError(f'Invalid name type - {name!r}.')
        _BaseName.__init__(self, name)
        self.__words = self._preprocess(name)

    def _repr(self):
        """
        Get the string representation of the character name.

        :returns: The string representation of the character name.
        :rtype: str
        """
        return self.__seperator__.join(self.__words)

    def __contains__(self, item: Union[str, List[str]]) -> bool:
        """
        Check if the character name contains a specific string or list of strings.

        :param item: The string or list of strings to check.
        :type item: Union[str, List[str]]
        :returns: True if the character name contains the specified string or list of strings, False otherwise.
        :rtype: bool
        """
        target = self._preprocess(item)
        n = len(target)
        for i in range(0, len(self.__words) - n + 1):
            if self.__words[i:i + n] == target:
                return True

        return False

    def __len__(self):
        """
        Calculate the length of the character name.

        :returns: The length of the character name.
        :rtype: int
        """
        return len(self.__words)

    def __getitem__(self, item) -> List[str]:
        """
        Get the item at the specified index or slice.

        :param item: The index or slice to retrieve.
        :type item: Union[int, slice]
        :returns: The item at the specified index or slice.
        :rtype: List[str]
        """
        return self.__words[item]

    _SPLITTERS = re.compile(r'[\s_,]+')

    @classmethod
    def _preprocess(cls, name: Union[str, List[str]]) -> List[str]:
        """
        Preprocess the character name.

        This method should be implemented by the subclasses to provide custom preprocessing logic.

        :param name: The name of the character.
        :type name: Union[str, List[str]]
        :returns: The preprocessed name.
        :rtype: List[str]
        """
        if isinstance(name, str):
            words = cls._SPLITTERS.split(name.strip().lower())
        else:
            words = name

        words = [wd.lower().strip() for wd in words]
        words = [wd for wd in words if wd]
        return words


class ChineseName(TextName):
    """
    Class for modeling Chinese character names.

    The class inherits from TextName and represents character names written in Chinese.

    Usage:
    - Use the ChineseName class to model Chinese character names.

    :param name: The name of the character.
    :type name: str
    """
    pass


class JapaneseName(TextName):
    """
    Class for modeling Japanese character names.

    The class inherits from TextName and represents character names written in Japanese.

    Usage:
    - Use the JapaneseName class to model Japanese character names.

    :param name: The name of the character.
    :type name: str
    """
    pass


class EnglishName(SegmentName):
    """
    Class for modeling English character names.

    The class inherits from SegmentName and represents character names written in English.

    Usage:
    - Use the EnglishName class to model English character names.

    :param name: The name of the character.
    :type name: Union[SegmentName, TextName, str, List[str]]
    """

    @classmethod
    def _word_trans(cls, name: str) -> str:
        """
        Translate individual words in the character name.

        This method should be implemented by the subclasses to provide custom word translation logic.

        :param name: The name of the character.
        :type name: str
        :returns: The translated name.
        :rtype: str
        """
        return name

    @classmethod
    def _preprocess(cls, name: Union[SegmentName, TextName, str, List[str]]) -> List[str]:
        """
        Preprocess the character name.

        This method should be implemented by the subclasses to provide custom preprocessing logic.

        :param name: The name of the character.
        :type name: Union[SegmentName, TextName, str, List[str]]
        :returns: The preprocessed name.
        :rtype: List[str]
        """
        if isinstance(name, SegmentName):
            return name[:]
        elif isinstance(name, TextName):
            return cls._preprocess(str(name))
        elif isinstance(name, str):
            name = cls._word_trans(name).replace(chr(160), ' ')
        elif isinstance(name, list):
            name = [cls._word_trans(wd) for wd in name]
        else:
            raise TypeError(f'Invalid name type - {name!r}.')

        return SegmentName._preprocess(name)


class GenericAliasName(TextName):
    """
    Class for modeling generic character names with aliases.

    The class inherits from TextName and represents generic character names that may have aliases.

    Usage:
    - Use the GenericAliasName class to model generic character names with aliases.

    :param name: The name of the character.
    :type name: str
    """

    @classmethod
    def _eqprocess(cls, name: str) -> str:
        """
        Preprocess the character name for equality comparison.

        This method should be implemented by the subclasses to provide custom preprocessing logic.

        :param name: The name of the character.
        :type name: str
        :returns: The preprocessed name for equality comparison.
        :rtype: str
        """
        return ' '.join(filter(bool, re.split(r'[\W_]+', TextName._eqprocess(name))))



import re
from enum import IntEnum, unique, Enum


@unique
class Rarity(IntEnum):
    """
    Enum representing the rarity of a character in Blue Archive.
    """

    R = 0x1
    SR = 0x2
    SSR = 0x3

    @classmethod
    def loads(cls, val) -> 'Rarity':
        """
        Load a rarity value.

        :param val: The rarity value to load.
        :type val: Union[int, :class:`Rarity`]
        :return: The loaded rarity.
        :rtype: :class:`Rarity`
        :raises ValueError: If the value is invalid.
        :raises TypeError: If the type is invalid.
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item

            raise ValueError(f'Invalid level value - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')


@unique
class WeaponType(Enum):
    """
    Enum representing the weapon type of a character in Blue Archive.
    """

    SMG = 'SMG'
    RL = 'RL'
    HG = 'HG'
    FT = 'FT'
    SG = 'SG'
    MT = 'MT'
    AR = 'AR'
    MG = 'MG'
    GL = 'GL'
    SR = 'SR'
    RG = 'RG'

    @classmethod
    def loads(cls, val) -> 'WeaponType':
        """
        Load a weapon type value.

        :param val: The weapon type value to load.
        :type val: Union[str, WeaponType]
        :return: The loaded weapon type.
        :rtype: WeaponType
        :raises ValueError: If the value is invalid.
        :raises TypeError: If the type is invalid.
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            pval = val.upper()
            if pval in cls.__members__:
                return cls.__members__[pval]

            raise ValueError(f'Invalid class value - {val!r}.')
        else:
            raise TypeError(f'Invalid class type - {val!r}.')

    def __eq__(self, other):
        if isinstance(other, WeaponType):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False


@unique
class Role(Enum):
    """
    Enum representing the role of a character in Blue Archive.
    """

    ATTACKER = 'ATTACKER'
    TACTICAL_SUPPORT = 'TACTICAL SUPPORT'
    HEALER = 'HEALER'
    SUPPORT = 'SUPPORT'
    TANK = 'TANK'

    @classmethod
    def loads(cls, val) -> 'Role':
        """
        Load a role value.

        :param val: The role value to load.
        :type val: Union[str, Role]
        :return: The loaded role.
        :rtype: Role
        :raises ValueError: If the value is invalid.
        :raises TypeError: If the type is invalid.
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            pval = re.sub(r'\W+', '_', val.upper()).strip('_')
            if pval in cls.__members__:
                return cls.__members__[pval]

            raise ValueError(f'Invalid class value - {val!r}.')
        else:
            raise TypeError(f'Invalid class type - {val!r}.')

    def __eq__(self, other):
        if isinstance(other, Role):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False


@unique
class AttackType(Enum):
    """
    Enum representing the attack type of a character in Blue Archive.
    """

    PENETRATION = 'PENETRATION'
    EXPLOSIVE = 'EXPLOSIVE'
    MYSTIC = 'MYSTIC'

    @classmethod
    def loads(cls, val) -> 'AttackType':
        """
        Load an attack type value.

        :param val: The attack type value to load.
        :type val: Union[str, AttackType]
        :return: The loaded attack type.
        :rtype: AttackType
        :raises ValueError: If the value is invalid.
        :raises TypeError: If the type is invalid.
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            pval = re.sub(r'\W+', '_', val.upper()).strip('_')
            if pval in cls.__members__:
                return cls.__members__[pval]

            raise ValueError(f'Invalid class value - {val!r}.')
        else:
            raise TypeError(f'Invalid class type - {val!r}.')

    def __eq__(self, other):
        if isinstance(other, AttackType):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False



import re
from typing import List, Tuple, Optional

from .name import ChineseName, JapaneseName, KoreanName, EnglishName
from .property import Rarity, Clazz, Burst
from ..base import Character as _BaseCharacter, Gender
from ..base.name import _BaseName


class Character(_BaseCharacter):
    __game_name__ = 'nikke'
    __official_name__ = 'Nikke: Goddess of Victory'
    __cnname_class__ = ChineseName
    __jpname_class__ = JapaneseName
    __enname_class__ = EnglishName

    def __init__(self, raw_data):
        self.__raw_data = raw_data

    def _index(self):
        return re.sub(r'[\W_]+', '_', self.__raw_data['enname'][0].lower()).strip('_')

    def _cnname(self):
        return self.__raw_data['cnname']

    def _jpname(self):
        return self.__raw_data['jpname'][0] if self.__raw_data['jpname'] else None

    def _jpnames(self):
        return self.__raw_data['jpname']

    def _enname(self):
        return self.__raw_data['enname'][0] if self.__raw_data['enname'] else None

    def _ennames(self):
        return self.__raw_data['enname']

    def _krname(self):
        return self.__raw_data['krname'][0] if self.__raw_data['krname'] else None

    @property
    def krname(self) -> Optional[KoreanName]:
        name = self._krname()
        return KoreanName(name) if name is not None else None

    def _krnames(self):
        return self.__raw_data['krname']

    @property
    def krnames(self) -> List[KoreanName]:
        return [KoreanName(name) for name in self._krnames()]

    def _names(self) -> List[_BaseName]:
        return [*_BaseCharacter._names(self), *self.krnames]

    def _gender(self):
        return Gender.FEMALE

    @property
    def rarity(self) -> Rarity:
        return Rarity.loads(self.__raw_data['rarity'])

    @property
    def clazz(self) -> str:
        return Clazz.loads(self.__raw_data['class'])

    @property
    def weapon_type(self) -> str:
        return self.__raw_data['weapontype']

    @property
    def manufacturer(self) -> str:
        return self.__raw_data['manufacturer']

    @property
    def burst(self) -> str:
        return Burst.loads(self.__raw_data['burst'])

    @property
    def code(self) -> str:
        return self.__raw_data['code']

    def _skins(self) -> List[Tuple[str, str]]:
        return [
            (skin['name'], skin['url'])
            for skin in self.__raw_data['skins']
        ]

    def _release_time(self):
        return self.__raw_data['release']['time']

    def _order(self):
        return self._release_time() or 0.0, 1 if self._is_extra() else 0

    def _is_extra(self) -> bool:
        return self._enname() and ':' in self._enname()

    def __repr__(self):
        return f'<{type(self).__name__} {"/".join(map(str, self._names()))}, ' \
               f'{self.rarity.name}({int(self.rarity)}{"*" * self.rarity}), ' \
               f'class: {self.clazz}, burst: {self.burst}, weapon_type: {self.weapon_type}>'



import re
from typing import List, Tuple

from .name import EnglishName, JapaneseName, ChineseName, ChineseAliasName
from .property import Rarity, WeaponType, Role, AttackType
from ..base import Character as _BaseCharacter, Gender


class Character(_BaseCharacter):
    """
    Represents a character in Blue Archive.
    Inherits from the base class _BaseCharacter.
    """
    __game_name__ = 'bluearchive'
    __official_name__ = 'Blue Archive'
    __cnname_class__ = ChineseName
    __enname_class__ = EnglishName
    __jpname_class__ = JapaneseName
    __alias_name_class__ = ChineseAliasName

    def __init__(self, raw_data: dict):
        """
        Initializes a Character object with raw data.

        :param raw_data: The raw data of the character.
        :type raw_data: dict
        """
        self.__raw_data = raw_data

    def _index(self) -> str:
        """
        Generates the index of the character.

        :return: The generated index.
        :rtype: str
        """
        return re.sub(r'\W+', '_', self._enname().lower())

    def _cnname(self):
        """
        Returns the Chinese name of the character.

        :return: The Chinese name.
        :rtype: str
        """
        return self.__raw_data['cnnames'][0]

    def _jpname(self):
        """
        Returns the Japanese name of the character.

        :return: The Japanese name.
        :rtype: str
        """
        return self.__raw_data['jpnames'][0]

    def _enname(self):
        """
        Returns the English name of the character.

        :return: The English name.
        :rtype: str
        """
        return self.__raw_data['ennames'][0]

    def _cnnames(self):
        """
        Returns a list of Chinese names of the character.

        :return: The list of Chinese names.
        :rtype: List[str]
        """
        return self.__raw_data['cnnames']

    def _jpnames(self):
        """
        Returns a list of Japanese names of the character.

        :return: The list of Japanese names.
        :rtype: List[str]
        """
        return self.__raw_data['jpnames']

    def _ennames(self):
        """
        Returns a list of English names of the character.

        :return: The list of English names.
        :rtype: List[str]
        """
        return self.__raw_data['ennames']

    def _gender(self):
        """
        Returns the gender of the character.

        :return: The gender.
        :rtype: Gender
        """
        return Gender.FEMALE

    @property
    def rarity(self) -> Rarity:
        """
        Returns the rarity of the character.

        :return: The rarity.
        :rtype: :class:`Rarity`
        """
        return Rarity.loads(self.__raw_data['data']['rarity'])

    @property
    def role(self) -> Role:
        """
        Returns the role of the character.

        :return: The role.
        :rtype: Role
        """
        return Role.loads(self.__raw_data['data']['role'])

    @property
    def weapon_type(self) -> WeaponType:
        """
        Returns the weapon type of the character.

        :return: The weapon type.
        :rtype: WeaponType
        """
        return WeaponType.loads(self.__raw_data['data']['weapon_type'])

    @property
    def attack_type(self) -> AttackType:
        """
        Returns the attack type of the character.

        :return: The attack type.
        :rtype: AttackType
        """
        return AttackType.loads(self.__raw_data['data']['attack_type'])

    def _skins(self) -> List[Tuple[str, str]]:
        """
        Returns the list of skins for the character.

        :return: The list of skins.
        :rtype: List[Tuple[str, str]]
        """
        return [(item['name'], item['url']) for item in self.__raw_data['skins']]

    def _release_time(self):
        """
        Returns the release time of the character.

        :return: The release time.
        :rtype: int
        """
        release_info = self.__raw_data['release']
        return release_info['time']

    def _order(self):
        """
        Returns the order of the character for sorting.

        :return: The order.
        :rtype: int
        """
        return self._release_time()

    def __repr__(self):
        """
        Returns a string representation of the character.

        :return: The string representation.
        :rtype: str
        """
        return f'<{type(self).__name__} {"/".join(map(str, self._names()))}, ' \
               f'{self.rarity.name}({int(self.rarity)}{"*" * self.rarity}), ' \
               f'role: {self.role}, attack: {self.attack_type}, weapon: {self.weapon_type}>'



import json
from functools import lru_cache
from typing import Union, Mapping, List, Tuple

from huggingface_hub import hf_hub_download

from gchar.games import get_character
from gchar.games.base import Character
from gchar.games.dispatch.access import GAME_CHARS
from .base import SITES

_AVAIL_SITE_NAMES = set(SITES.keys())


@lru_cache()
def _get_tags_for_game_site(game: str, site: str):
    """
    Get the tags for a specific game and site.

    :param game: The name of the game.
    :type game: str
    :param site: The name of the site.
    :type site: str
    :return: The tags for the game and site.
    :rtype: List[dict]
    """
    with open(hf_hub_download(
            GAME_CHARS[game].__repository__,
            filename=f'{game}/tags_{SITES[site]}.json',
            repo_type='dataset'
    ), 'r', encoding='utf-8') as f:
        return json.load(f)['data']


@lru_cache()
def _get_tag_mapping(game: str, site: str) -> Mapping[Union[str, int], List[dict]]:
    """
    Get the mapping of tags for a specific game and site.

    :param game: The name of the game.
    :type game: str
    :param site: The name of the site.
    :type site: str
    :return: The mapping of tags.
    :rtype: Mapping[Union[str, int], List[dict]]
    """
    tags = _get_tags_for_game_site(game, site)
    return {tag['index']: tag['tags'] for tag in tags}


def _get_raw_site_tags(ch: Union[Character, str], site: str, sure_only: bool = False,
                       allow_fuzzy: bool = False, **kwargs) -> List[dict]:
    """
    Get the raw tags for a character and site.

    :param ch: The character instance or name.
    :type ch: Union[Character, str]
    :param site: The name of the site.
    :type site: str
    :param sure_only: Whether to include only sure tags.
    :type sure_only: bool
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param kwargs: Additional arguments for character retrieval.
    :return: The raw tags for the character and site.
    :rtype: List[dict]
    """
    if site not in _AVAIL_SITE_NAMES:
        raise ValueError(f'Unsupported site - {site!r}.\n'
                         f'{list(SITES.keys())!r} are supported.')
    if not isinstance(ch, Character):
        ch = get_character(ch, allow_fuzzy, **kwargs)

    _tag_mapping = _get_tag_mapping(ch.__game_name__, site)
    if ch.index in _tag_mapping:
        tags = _tag_mapping[ch.index]
        if sure_only:
            tags = [tag for tag in tags if tag['sure']]
        return tags
    else:
        return []


def list_site_tags(ch: Union[Character, str], site: str, sure_only: bool = False,
                   with_posts: bool = False, allow_fuzzy: bool = False, **kwargs) \
        -> Union[List[str], List[Tuple[str, int]]]:
    """
    List the tags for a character and site.

    :param ch: The character instance or name.
    :type ch: Union[Character, str]
    :param site: The name of the site.
    :type site: str
    :param sure_only: Whether to include only sure tags.
    :type sure_only: bool
    :param with_posts: Whether to include the number of posts for each tag.
    :type with_posts: bool
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param kwargs: Additional arguments for character retrieval.
    :return: The list of tags for the character and site.
    :rtype: Union[List[str], List[Tuple[str, int]]]

    Examples::
        >>> from gchar.resources.sites import list_site_tags
        >>>
        >>> list_site_tags('amiya', 'danbooru')
        ['amiya_(arknights)', 'amiya_(guard)_(arknights)', 'amiya_(newsgirl)_(arknights)', 'amiya_(fresh_fastener)_(arknights)', 'amiya_(planter)_(arknights)', 'amiya_(guard)_(touch_the_stars)_(arknights)']
        >>> list_site_tags('surtr', 'danbooru')
        ['surtr_(arknights)', 'surtr_(colorful_wonderland)_(arknights)', 'surtr_(liberte_echec)_(arknights)']
        >>> list_site_tags('surtr', 'danbooru', with_posts=True)
        [('surtr_(arknights)', 2666), ('surtr_(colorful_wonderland)_(arknights)', 483), ('surtr_(liberte_echec)_(arknights)', 306)]
        >>> list_site_tags('surtr', 'zerochan')
        ['Surtr (Arknights)']
        >>> list_site_tags('surtr', 'anime_pictures')
        ['surtr (arknights)', 'surtr (liberte echec) (arknights)', 'surtr (colorful wonderland) (arknights)']
        >>> list_site_tags('surtr', 'anime_pictures', with_posts=True)
        [('surtr (arknights)', 471), ('surtr (liberte echec) (arknights)', 75), ('surtr (colorful wonderland) (arknights)', 49)]
    """
    tags = _get_raw_site_tags(ch, site, sure_only, allow_fuzzy, **kwargs)
    if with_posts:
        return [(tag['name'], tag['count']) for tag in tags]
    else:
        return [tag['name'] for tag in tags]


def get_site_tag(ch: Union[Character, str], site: str, sure_only: bool = False,
                 with_posts: bool = False, allow_fuzzy: bool = False, **kwargs) \
        -> Union[str, Tuple[str, int], None]:
    """
    Get a single tag for a character and site.

    :param ch: The character instance or name.
    :type ch: Union[Character, str]
    :param site: The name of the site.
    :type site: str
    :param sure_only: Whether to include only sure tags.
    :type sure_only: bool
    :param with_posts: Whether to include the number of posts for the tag.
    :type with_posts: bool
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param kwargs: Additional arguments for character retrieval.
    :return: The tag for the character and site.
    :rtype: Union[str, Tuple[str, int], None]

    Examples:
        >>> from gchar.resources.sites import get_site_tag
        >>>
        >>> get_site_tag('amiya', 'danbooru')
        'amiya_(arknights)'
        >>> get_site_tag('surtr', 'danbooru')
        'surtr_(arknights)'
        >>> get_site_tag('surtr', 'danbooru', with_posts=True)
        ('surtr_(arknights)', 2666)
        >>> get_site_tag('surtr', 'zerochan')
        'Surtr (Arknights)'
        >>> get_site_tag('surtr', 'anime_pictures')
        'surtr (arknights)'
        >>> get_site_tag('surtr', 'anime_pictures', with_posts=True)
        ('surtr (arknights)', 471)
    """
    tags = list_site_tags(ch, site, sure_only, with_posts, allow_fuzzy, **kwargs)
    if tags:
        return tags[0]
    else:
        return None



import json
from dataclasses import dataclass
from functools import lru_cache
from typing import Union

from huggingface_hub import HfFileSystem, hf_hub_url

hf_fs = HfFileSystem()


@lru_cache()
def _read_hf_skins_meta(game: str, ch_index: Union[int, str]):
    """
    Read the metadata for the skins of a game character from the Hugging Face Hub.

    The function reads the metadata file for the specified game character from the Hugging Face Hub dataset repository.
    It retrieves the names and URLs of the skins that have been synchronized to the Hugging Face Hub.

    :param game: The game name.
    :type game: str
    :param ch_index: The index or ID of the game character.
    :type ch_index: Union[int, str]
    :returns: A dictionary mapping skin names to their Hugging Face Hub URLs.
    :rtype: dict
    """
    from gchar.games.dispatch.access import GAME_CHARS
    repository = GAME_CHARS[game].__repository__

    online_dir = f'datasets/{repository}/{game}/{ch_index}'
    meta_file = f'{online_dir}/.meta.json'
    if hf_fs.exists(meta_file):
        raw = json.loads(hf_fs.read_text(meta_file))
        return {
            skin['metadata']['name']: hf_hub_url(
                repository,
                filename=f'{game}/{ch_index}/{skin["name"]}',
                repo_type='dataset'
            ) for skin in raw['files']
        }
    else:
        return {}


@dataclass
class Skin:
    """
    Represents a skin of a game character.

    The class stores information about the game, character index, skin name, and original URL of the skin.
    It provides a property `selected_url` that returns the URL of the skin, which is either the Hugging Face Hub URL
    if the skin has been synchronized or the original URL.

    :param game: The game name.
    :type game: str
    :param ch_index: The index or ID of the game character.
    :type ch_index: Union[int, str]
    :param name: The name of the skin.
    :type name: str
    :param url: The original URL of the skin.
    :type url: str
    """

    game: str
    ch_index: Union[int, str]
    name: str
    url: str

    @property
    def selected_url(self) -> str:
        """
        Get the URL of the skin.

        If the skin has been synchronized to the Hugging Face Hub, the Hugging Face Hub URL is returned.
        Otherwise, the original URL is returned.

        :returns: The URL of the skin.
        :rtype: str
        """
        return _read_hf_skins_meta(self.game, self.ch_index).get(self.name, self.url)



import re

from unidecode import unidecode

from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName


class ChineseName(_GenericChineseName):
    """
    Represents the Chinese name of a character in Blue Archive.
    Inherits from the base class _GenericChineseName.
    """
    pass


class ChineseAliasName(_GenericChineseName):
    """
    Represents the Chinese alias name of a character in Blue Archive.
    Inherits from the base class _GenericChineseName.
    """
    pass


class JapaneseName(_GenericJapaneseName):
    """
    Represents the Japanese name of a character in Blue Archive.
    Inherits from the base class _GenericJapaneseName.
    """
    pass


class EnglishName(_GenericEnglishName):
    """
    Represents the English name of a character in Blue Archive.
    Inherits from the base class _GenericEnglishName.
    """

    @classmethod
    def _word_trans(cls, text: str):
        """
        Translates the text into English words.

        :param text: The text to translate.
        :type text: str
        :return: The translated English words.
        :rtype: str
        """
        return ' '.join(re.findall('[a-zA-Z0-9]+', unidecode(text.lower())))



from transliterate import translit

from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName


class ChineseName(_GenericChineseName):
    """
    A class for modeling Chinese names of characters in the Arknights game.
    """

    @classmethod
    def _preprocess(cls, name: str) -> str:
        """
        Preprocess the Chinese name by replacing '・' with '·'.

        :param name: The input Chinese name.
        :type name: str
        :returns: The preprocessed Chinese name.
        :rtype: str
        """
        return _GenericChineseName._preprocess(name).replace('・', '·')


class ChineseAliasName(_GenericChineseName):
    """
    A class for modeling Chinese alias names of characters in the Arknights game.
    """

    @classmethod
    def _preprocess(cls, name: str) -> str:
        """
        Preprocess the Chinese alias name by replacing '・' with '·'.

        :param name: The input Chinese alias name.
        :type name: str
        :returns: The preprocessed Chinese alias name.
        :rtype: str
        """
        return _GenericChineseName._preprocess(name).replace('・', '·')


class JapaneseName(_GenericJapaneseName):
    """
    A class for modeling Japanese names of characters in the Arknights game.
    """


class EnglishName(_GenericEnglishName):
    """
    A class for modeling English names of characters in the Arknights game.
    """

    @classmethod
    def _word_trans(cls, text: str):
        """
        Transliterate the English name using the 'ru' transliteration scheme.

        :param text: The input English name.
        :type text: str
        :returns: The transliterated English name.
        :rtype: str
        """
        return translit(text, 'ru', reversed=True)



import re

from unidecode import unidecode

from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName


class ChineseName(_GenericChineseName):
    """
    A class representing Chinese names of characters in the Azur Lane game.
    Inherits from _GenericChineseName.
    """


class ChineseAliasName(_GenericChineseName):
    """
    A class representing Chinese alias names of characters in the Azur Lane game.
    Inherits from _GenericChineseName.
    """


class JapaneseName(_GenericJapaneseName):
    """
    A class representing Japanese names of characters in the Azur Lane game.
    Inherits from _GenericJapaneseName.
    """


class EnglishName(_GenericEnglishName):
    """
    A class representing English names of characters in the Azur Lane game.
    Inherits from _GenericEnglishName.
    """

    @classmethod
    def _word_trans(cls, text: str):
        """
        Perform word-based transliteration on the given text.

        :param text: The text to transliterate.
        :type text: str
        :return: The transliterated text.
        :rtype: str
        """
        return ' '.join(re.findall('[a-zA-Z0-9]+', unidecode(text.lower())))



import os
from functools import lru_cache, wraps


def optional_lru_cache(*pargs, **pkwargs):
    """
    Decorator that optionally applies LRU caching to a function's return value based on the presence of the 'NO_CACHE'
    environment variable.

    The function is decorated with the `lru_cache` decorator with the provided arguments `pargs` and `pkwargs` to enable
    caching. However, if the 'NO_CACHE' environment variable is present, the original function is called directly
    without caching.

    :param pargs: Positional arguments to be passed to the `lru_cache` decorator.
    :type pargs: tuple
    :param pkwargs: Keyword arguments to be passed to the `lru_cache` decorator.
    :type pkwargs: dict
    :returns: Decorated function.
    :rtype: callable
    """

    def _decorator(func):
        _cached_func = lru_cache(*pargs, **pkwargs)(func)

        @wraps(func)
        def _new_func(*args, **kwargs):
            if os.environ.get('NO_CACHE', None):
                return func(*args, **kwargs)
            else:
                return _cached_func(*args, **kwargs)

        return _new_func

    return _decorator



from enum import unique, Enum, IntEnum


@unique
class Rarity(IntEnum):
    FOUR = 0x4
    FIVE = 0x5

    @classmethod
    def loads(cls, val) -> 'Rarity':
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item.value

            raise ValueError(f'Invalid level value - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')


@unique
class Weapon(Enum):
    CATALYST = 0x1
    CLAYMORE = 0x2
    SWORD = 0x3
    BOW = 0x4
    POLEARM = 0x5

    def __eq__(self, other):
        if isinstance(other, Weapon):
            return other.value == self.value
        else:
            try:
                other = Weapon.loads(other)
            except (TypeError, ValueError):
                return False
            else:
                return other == self

    @classmethod
    def loads(cls, val) -> 'Weapon':
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            pval = val.upper().strip()
            if pval in cls.__members__:
                return cls.__members__[pval]

            if pval in {'单手剑'}:
                return cls.SWORD
            elif pval in {'双手剑'}:
                return cls.CLAYMORE
            elif pval in {'长柄武器'}:
                return cls.POLEARM
            elif pval in {'法器'}:
                return cls.CATALYST
            elif pval in {'弓'}:
                return cls.BOW

            raise ValueError(f'Invalid weapon value - {val!r}.')
        else:
            raise TypeError(f'Invalid weapon type - {val!r}.')


@unique
class Element(Enum):
    PYRO = 0x1
    ANEMO = 0x2
    CRYO = 0x3
    ELECTRO = 0x4
    DENDRO = 0x5
    GEO = 0x6
    HYDRO = 0x7

    def __eq__(self, other):
        if isinstance(other, Element):
            return other.value == self.value
        else:
            try:
                other = Element.loads(other)
            except (TypeError, ValueError):
                return False
            else:
                return other == self

    @classmethod
    def loads(cls, val) -> 'Element':
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            pval = val.upper().strip()
            if pval in cls.__members__:
                return cls.__members__[pval]

            if pval in {'火'}:
                return cls.PYRO
            elif pval in {'风'}:
                return cls.ANEMO
            elif pval in {'冰'}:
                return cls.CRYO
            elif pval in {'雷'}:
                return cls.ELECTRO
            elif pval in {'草'}:
                return cls.DENDRO
            elif pval in {'岩'}:
                return cls.GEO
            elif pval in {'水'}:
                return cls.HYDRO

            raise ValueError(f'Invalid element value - {val!r}.')
        else:
            raise TypeError(f'Invalid element type - {val!r}.')



import re
import warnings
from itertools import chain
from typing import Iterable, Iterator, Union, List, Tuple, Type, Mapping, Optional

from .base import _GAMES
from .keyword import _load_pixiv_names_for_game, _load_pixiv_alias_for_game
from ...games import get_character
from ...games.base import Character
from ...utils import optional_lru_cache


def _yield_tags(tags: Union[Tuple[str], List[str], str]) -> Iterator[str]:
    """
    Yield individual tags from a string or a list of tags.

    :param tags: The tags to yield.
    :type tags: Union[Tuple[str], List[str], str]
    :returns: An iterator of individual tags.
    :rtype: Iterator[str]
    """
    if isinstance(tags, str):
        for item in re.split(r'\s+', tags):
            if item:
                yield item
    elif isinstance(tags, (list, tuple, set)):
        for item in tags:
            yield from _yield_tags(item)


def _format_tags(positive, negative, or_clause=None):
    """
    Format the positive and negative tags into a valid Pixiv search tag string.

    :param positive: The positive tags.
    :type positive: Union[Tuple[str], List[str], str]
    :param negative: The negative tags.
    :type negative: Union[Tuple[str], List[str], str]
    :param or_clause: The OR clause tags.
    :type or_clause: Optional[Union[Tuple[str], List[str], str]]
    :returns: The formatted Pixiv search tag string.
    :rtype: str
    """
    positive_tags = list(_yield_tags(positive))
    negative_tags = list(_yield_tags(negative))
    or_clause_tags = list(_yield_tags(or_clause or ''))
    if len(or_clause_tags) == 1:
        or_tag = or_clause_tags[0]
        if or_tag not in positive_tags:
            positive_tags.append(or_tag)
        or_clause_tags = []

    positive_tags = set(positive_tags)
    or_clause_tags = set(or_clause_tags)
    negative_tags = set(negative_tags) - positive_tags - or_clause_tags
    positive_tags = sorted(positive_tags)
    negative_tags = sorted(negative_tags)
    or_clause_tags = sorted(or_clause_tags)

    all_phrases = [*positive_tags]
    if or_clause_tags:
        all_phrases.append(f'({" OR ".join(or_clause_tags)})')
    if negative_tags:
        all_phrases.extend((f'-{x}' for x in negative_tags))

    return ' '.join(all_phrases)


PIXIV_TAG_MAX_LENGTH = 256


class PixivCharPool:
    def __init__(self, chars: Iterable[Character],
                 names_dict: Mapping[str, Tuple[int, float, List[Tuple[str, int]]]],
                 names_alias: Mapping[Union[str, int], List[str]]):
        """
        Initialize the Pixiv character pool.

        :param chars: The iterable of Character instances.
        :type chars: Iterable[Character]
        :param names_dict: The mapping of character names to their Pixiv data.
        :type names_dict: Mapping[str, Tuple[int, float, List[Tuple[str, int]]]]
        :param names_alias: The mapping of character IDs or names to their Pixiv aliases.
        :type names_alias: Mapping[Union[str, int], List[str]]
        """
        self.__chars = list(chars)
        self.__names_dict = names_dict
        self.__names_alias = names_alias
        self.__all_names = sorted(
            set(self.__names_dict.keys()) |
            set(chain(*self.__names_alias.values()))
        )

    def __get_name_item(self, name) -> Optional[Tuple[int, float, List[Tuple[str, int]]]]:
        """
        Get the Pixiv data for a specific character name.

        :param name: The character name.
        :type name: str
        :returns: The Pixiv data tuple.
        :rtype: Optional[Tuple[int, float, List[Tuple[str, int]]]]
        """
        return self.__names_dict.get(name, None)

    def __get_name_count(self, name) -> int:
        """
        Get the count of illustrations for a specific character name.

        :param name: The character name.
        :type name: str
        :returns: The count of illustrations.
        :rtype: int
        """
        tpl = self.__get_name_item(name)
        if tpl:
            count, _, _ = tpl
            return count
        else:
            return 0

    def __get_name_pollution_ratio(self, name) -> float:
        """
        Get the pollution ratio for a specific character name.

        :param name: The character name.
        :type name: str
        :returns: The pollution ratio.
        :rtype: float
        """
        tpl = self.__get_name_item(name)
        if tpl:
            _, ratio, _ = tpl
            return ratio
        else:
            return 0.0

    def __get_name_pollution_words(self, name) -> List[Tuple[str, int]]:
        """
        Get the pollution words and their counts for a specific character name.

        :param name: The character name.
        :type name: str
        :returns: A list of pollution word-count tuples.
        :rtype: List[Tuple[str, int]]
        """
        tpl = self.__get_name_item(name)
        if tpl:
            _, _, pollution = tpl
            return pollution
        else:
            return []

    def _iter_dup_names(self, name: str) -> Iterator[str]:
        """
        Iterate over duplicate names that contain the given name.

        :param name: The name to search for.
        :type name: str
        :returns: An iterator of duplicate names.
        :rtype: Iterator[str]
        """
        for sname in self.__all_names:
            if name != sname and name in sname:
                yield sname

    def get_tag(self, char: Character, use_english: bool = False, positive=None, negative=None,
                max_exclude_per_word: int = 20, max_exclude: int = 20, max_pollution_ratio: float = 0.8,
                max_length: int = PIXIV_TAG_MAX_LENGTH):
        """
        Generate a Pixiv search tag for a specific character.

        :param char: The character instance or name.
        :type char: Union[Character, str]
        :param use_english: Whether to use English names in the tag.
        :type use_english: bool
        :param positive: The positive tags to include.
        :type positive: Optional[Union[Tuple[str], List[str], str]]
        :param negative: The negative tags to exclude.
        :type negative: Optional[Union[Tuple[str], List[str], str]]
        :param max_exclude_per_word: The maximum number of excluded tags per word.
        :type max_exclude_per_word: int
        :param max_exclude: The maximum number of excluded tags.
        :type max_exclude: int
        :param max_pollution_ratio: The maximum pollution ratio for including tags.
        :type max_pollution_ratio: float
        :param max_length: The maximum length of the generated tag.
        :type max_length: int
        :returns: The generated Pixiv search tag.
        :rtype: str
        """
        if not isinstance(char, Character):
            raise TypeError(f'Invalid character type - {char!r}.')

        char_names = [*char.cnnames, *char.jpnames]
        if use_english:
            char_names.extend(char.ennames)
        if char.index in self.__names_alias:
            char_names.extend(self.__names_alias[char.index])

        char_names = sorted(set(map(lambda x: str(x).lower(), char_names)))

        origin_positive = positive
        origin_negative = negative
        positive = set(_yield_tags(positive or []))
        negative = set(_yield_tags(negative or [])) - positive
        or_clause = set()

        exclude_names = set()
        exclude_name_pairs = []
        min_pollution = 1.0
        for chname in char_names:
            s_chname = str(chname)
            name_pollution_ratio = self.__get_name_pollution_ratio(s_chname)
            min_pollution = min(name_pollution_ratio, min_pollution)
            all_exnames = list(self._iter_dup_names(s_chname))
            if name_pollution_ratio <= max_pollution_ratio and len(all_exnames) <= max_exclude_per_word:
                or_clause.add(s_chname)

                for pword, pcnt in self.__get_name_pollution_words(s_chname):
                    if pword not in positive and pword not in or_clause and \
                            pword != char and pword not in exclude_names:
                        exclude_names.add(pword)
                        exclude_name_pairs.append((pword, pcnt, 1))

                for exname in all_exnames:
                    if exname not in positive and exname not in or_clause and \
                            exname != char and exname not in exclude_names:
                        exclude_names.add(exname)
                        exclude_name_pairs.append((exname, self.__get_name_count(exname), 0))

        if or_clause:
            exclude_name_pairs = sorted(exclude_name_pairs, key=lambda x: (x[2], -x[1], len(x[0]), x[0]))[:max_exclude]

            while True:
                current_negative = set(negative)
                for exname, _, _ in exclude_name_pairs:
                    current_negative.add(exname)

                ret_keyword = _format_tags(positive, current_negative, or_clause)
                if len(ret_keyword) > max_length:
                    exclude_name_pairs = exclude_name_pairs[:-1]
                else:
                    return ret_keyword
        else:
            return self.get_tag(char, use_english, origin_positive, origin_negative,
                                max_exclude_per_word, max_exclude, max_pollution_ratio=min_pollution + 0.015)

    def _iter_end_dup_names(self, name: str) -> Iterator[str]:
        """
        Iterate over names that end with the given name.

        :param name: The name to search for.
        :type name: str
        :returns: An iterator of names.
        :rtype: Iterator[str]
        """
        for sname in self.__all_names:
            if name != sname and sname.endswith(name):
                yield sname

    def get_simple_tag(self, char: Character, base_tag: str, max_exclude: int = 20):
        """
        Generate a simplified Pixiv search tag for a specific character.

        :param char: The character instance.
        :type char: Character
        :param base_tag: The base tag to append to the character name.
        :type base_tag: str
        :param max_exclude: The maximum number of excluded tags.
        :type max_exclude: int
        :returns: The generated simplified Pixiv search tag.
        :rtype: str
        """
        if not isinstance(char, Character):
            raise TypeError(f'Invalid character type - {char!r}.')

        positive = set()
        negative = set()
        or_clause = set()
        if char.jpnames:
            exclude_names = set()
            for jpname in char.jpnames:
                positive.add(f'{jpname}({base_tag})')
                for exname in self._iter_end_dup_names(str(jpname)):
                    exclude_names.add(exname)

            for exname in sorted(exclude_names, key=lambda x: self.__get_name_count(x), reverse=True)[:max_exclude]:
                negative.add(exname)

        else:
            raise ValueError(f'Japanese name not found for character - {char!r}.')

        return _format_tags(positive, negative, or_clause)


@optional_lru_cache()
def _get_char_pool(cls: Type[Character], **kwargs):
    """
    Get the Pixiv character pool for a specific character class.

    :param cls: The character class.
    :type cls: Type[Character]
    :returns: The PixivCharPool instance.
    :rtype: PixivCharPool
    """
    names_dict = _load_pixiv_names_for_game(cls)
    names_alias = _load_pixiv_alias_for_game(cls)
    return PixivCharPool(cls.all(**kwargs), names_dict, names_alias)


def get_pixiv_keywords(char: Union[Character, str], simple: bool = False, use_english: bool = True,
                       includes=None, exclude=None,
                       allow_fuzzy: bool = True, fuzzy_threshold: int = 70, max_exclude: int = 20,
                       max_pollution_ratio: float = 0.8, max_length: int = PIXIV_TAG_MAX_LENGTH, **kwargs):
    """
    Get the Pixiv search keywords for a specific character.

    :param char: The character instance or name.
    :type char: Union[Character, str]
    :param simple: Whether to generate a simplified tag.
    :type simple: bool
    :param use_english: Whether to use English names in the tag.
    :type use_english: bool
    :param includes: The positive tags to include.
    :type includes: Optional[Union[Tuple[str], List[str], str]]
    :param exclude: The negative tags to exclude.
    :type exclude: Optional[Union[Tuple[str], List[str], str]]
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param fuzzy_threshold: The threshold for fuzzy matching.
    :type fuzzy_threshold: int
    :param max_exclude: The maximum number of excluded tags.
    :type max_exclude: int
    :param max_pollution_ratio: The maximum pollution ratio for including tags.
    :type max_pollution_ratio: float
    :param max_length: The maximum length of the generated tag.
    :type max_length: int
    :returns: The generated Pixiv search keywords.
    :rtype: str
    :raises ValueError: If the character is unknown.

    Examples::
        >>> from gchar.resources.pixiv import get_pixiv_keywords
        >>>
        >>> get_pixiv_keywords('amiya')
        'アークナイツ (amiya OR アーミヤ OR 阿米娅)'
        >>> get_pixiv_keywords('surtr')
        'アークナイツ (surtr OR スルト OR 史尔特尔) -明日方舟スルト'
        >>> get_pixiv_keywords('dusk')  # ケルシー and ドロシー way cause search noises
        'アークナイツ (dusk OR シー OR 夕) -ケルシー -シージ -シーン -ドロシー -ルーシー -夕張 -夕日 -夕焼け'
    """
    original_char = char
    if not isinstance(char, Character):
        char = get_character(char, allow_fuzzy, fuzzy_threshold, **kwargs)
    if not char:
        raise ValueError(f'Unknown character - {original_char!r}.')
    if max_length > PIXIV_TAG_MAX_LENGTH:
        warnings.warn(UserWarning(f'The maximum length pixiv supports is {PIXIV_TAG_MAX_LENGTH}, '
                                  f'but {max_length!r} is given. '
                                  f'This may result in no search results.'), stacklevel=2)

    pool = _get_char_pool(type(char), **kwargs)
    game_tag, base_tag = _GAMES[char.__game_name__]

    try:
        if simple:
            return pool.get_simple_tag(char, base_tag, max_exclude=max_exclude)
    except ValueError:
        warnings.warn(UserWarning(f'No japanese name for {char!r}, falling back to full tag.'), stacklevel=2)

    return pool.get_tag(char, use_english, positive=[includes, game_tag], negative=exclude,
                        max_exclude=max_exclude, max_pollution_ratio=max_pollution_ratio, max_length=max_length)



from typing import List, Tuple, Mapping, Optional

from .name import EnglishName, ChineseName, ChineseAliasName, JapaneseName
from .property import Clazz, Rarity
from ..base import Character as _BaseCharacter
from ..girlsfrontline import Character as GFCharacter
from ...utils import optional_lru_cache


class Character(_BaseCharacter):
    __game_name__ = 'neuralcloud'
    __official_name__ = 'Neural Cloud'
    __enname_class__ = EnglishName
    __cnname_class__ = ChineseName
    __jpname_class__ = JapaneseName
    __alias_name_class__ = ChineseAliasName

    def __init__(self, raw_data: dict):
        self.__raw_data = raw_data

    def _index(self) -> int:
        return self.__raw_data['id']

    def _cnname(self):
        return self.__raw_data['cnname']

    @property
    def cnnames(self):
        names = [self.__cnname_class__(name) for name in self._cnnames() if name]
        if self.gf_char:
            names.extend(self.gf_char.cnnames)

        return [name for name in names if name]

    def _jpname(self):
        return self.__raw_data['jpname']

    @property
    def jpnames(self):
        names = [self.__jpname_class__(name) for name in self._jpnames() if name]
        if self.gf_char:
            names.extend(self.gf_char.jpnames)
        return [name for name in names if name]

    def _enname(self):
        return self.__raw_data['enname']

    @property
    def ennames(self):
        names = [self.__enname_class__(name) for name in self._ennames() if name]
        if self.gf_char:
            names.extend(self.gf_char.ennames)
        return [name for name in names if name]

    def _custom_alias_names(self):
        return self.__raw_data['alias']

    def _gender(self):
        return self.__raw_data['gender']

    @property
    def clazz(self) -> Clazz:
        return Clazz.loads(self.__raw_data['class'])

    @property
    def rarity(self) -> Rarity:
        return Rarity.loads(self.__raw_data['rarity'])

    @property
    def company(self) -> str:
        return self.__raw_data.get('company', None)

    @property
    def gf_char(self) -> Optional[GFCharacter]:
        if self.__raw_data["gf"]:
            return self._list_index_of_gf()[self.__raw_data["gf"]["id"]]
        else:
            return None

    def _skins(self) -> List[Tuple[str, str]]:
        return [(item['name'], item['url']) for item in self.__raw_data['skins']]

    def __repr__(self):
        # return f'233 {type(self).__name__} {self.index} {self._names()} {self.rarity} {self.clazz}'
        return f'<{type(self).__name__} {self.index} - {"/".join(map(str, self._names()))}, ' \
               f'rarity: {self.rarity}{"*" * self.rarity}, clazz: {self.clazz.name}>'

    def _release_time(self):
        release_info = self.__raw_data['release']
        return release_info['time']

    def _order(self):
        return self._release_time()

    @classmethod
    @optional_lru_cache()
    def _list_index_of_gf(cls) -> Mapping[int, GFCharacter]:
        return {ch.index: ch for ch in GFCharacter.all(contains_extra=False)}



import re
from enum import IntEnum, Enum
from functools import total_ordering


class Rarity(IntEnum):
    R = 1
    SR = 2
    SSR = 3

    @classmethod
    def loads(cls, obj):
        if isinstance(obj, Rarity):
            return obj
        elif isinstance(obj, str):
            return cls.__members__[obj.upper().strip()]
        elif isinstance(obj, int):
            for name, item in cls.__members__.items():
                if item.value == obj:
                    return item

            raise ValueError(f'Unknown rarity value - {obj!r}.')
        else:
            raise TypeError(f'Unknown rarity type - {obj!r}.')


@total_ordering
class Clazz(Enum):
    ATTACKER = (('attackers', 'attacker', '火力型'), 1)
    DEFENDER = (('defenders', 'defender', '防御型'), 2)
    SUPPORTER = (('supporters', 'supporter', '辅助型'), 3)

    def __init__(self, aliases, number):
        self.aliases = aliases
        self.number = number

    @classmethod
    def loads(cls, obj):
        if isinstance(obj, Clazz):
            return obj
        elif isinstance(obj, str):
            for key, item in cls.__members__.items():
                names = set(map(str.lower, [key, *item.aliases]))
                if obj.lower().strip() in names:
                    return item
            raise ValueError(f'Unknown class - {obj!r}')
        elif isinstance(obj, int):
            for key, item in cls.__members__.items():
                if item.number == obj:
                    return item
            raise ValueError(f'Unknown class number - {obj!r}.')
        else:
            raise TypeError(f'Unknown class type - {obj!r}.')

    def __eq__(self, other):
        if isinstance(other, Clazz):
            return self.name == other.name
        else:
            try:
                return self == self.loads(other)
            except (KeyError, ValueError, TypeError):
                return False

    def __hash__(self):
        return hash(self.aliases)

    def __lt__(self, other):
        return self.number < other.number

    def __repr__(self):
        return f'<{self.__class__.__name__}.{self.name}: {self.number}>'


@total_ordering
class Burst(Enum):
    I = (('burst skill i', 'i'), 1)
    II = (('burst skill ii', 'ii'), 2)
    III = (('burst skill iii', 'iii'), 3)

    def __init__(self, aliases, number):
        self.aliases = aliases
        self.number = number

    @classmethod
    def _name_preprocess(cls, name):
        return re.sub(r'[\W_]+', '', name.lower().strip())

    @classmethod
    def loads(cls, obj):
        if isinstance(obj, Burst):
            return obj
        elif isinstance(obj, str):
            for key, item in cls.__members__.items():
                names = set(map(cls._name_preprocess, [key, *item.aliases]))
                if cls._name_preprocess(obj) in names:
                    return item
            raise ValueError(f'Unknown class - {obj!r}')
        elif isinstance(obj, int):
            for key, item in cls.__members__.items():
                if item.number == obj:
                    return item
            raise ValueError(f'Unknown class number - {obj!r}.')
        else:
            raise TypeError(f'Unknown class type - {obj!r}.')

    def __eq__(self, other):
        if isinstance(other, Burst):
            return self.name == other.name
        else:
            try:
                return self == self.loads(other)
            except (KeyError, ValueError, TypeError):
                return False

    def __hash__(self):
        return hash(self.aliases)

    def __lt__(self, other):
        return self.number < other.number

    def __repr__(self):
        return f'<{self.__class__.__name__}.{self.name}: {self.number}>'



from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName
from ..base import TextName


class ChineseName(_GenericChineseName):
    pass


class JapaneseName(_GenericJapaneseName):
    pass


class KoreanName(TextName):
    pass


class EnglishName(_GenericEnglishName):
    pass



import builtins
import json
from functools import lru_cache
from itertools import chain
from typing import List, Union, Type, Iterator, Optional, Tuple, Dict

from huggingface_hub import hf_hub_download, HfFileSystem

from .name import _BaseName, ChineseName, EnglishName, JapaneseName, GenericAliasName
from .property import Gender
from .skin import Skin
from ...utils import Comparable

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
import yaml

hf_fs = HfFileSystem()


class Character(Comparable):
    """
    Base class for modeling characters in different games.

    The class provides methods for accessing and managing character attributes, such as name, gender, skins, etc.
    It also provides class methods for retrieving and listing characters.

    Usage:
    - Inherit from the gchar.games.base.Character class and implement the required methods to customize the character model.

    :var __repository__: The repository name for the character data.
    :var __game_name__: The name of the game.
    :var __official_name__: The official name of the character.
    :var __cnname_class__: The class for modeling Chinese character names.
    :var __enname_class__: The class for modeling English character names.
    :var __jpname_class__: The class for modeling Japanese character names.
    :var __alias_name_class__: The optional class for modeling alias names.
    """

    __repository__: str = 'deepghs/game_characters'
    __game_name__: str
    __official_name__: str
    __cnname_class__: Type[ChineseName] = ChineseName
    __enname_class__: Type[EnglishName] = EnglishName
    __jpname_class__: Type[JapaneseName] = JapaneseName
    __alias_name_class__: Optional[Type[ChineseName]] = None

    def _index(self):
        """
        Get the index of the character.

        This method should be implemented by the subclasses.

        :returns: The index of the character.
        :rtype: Any
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def index(self):
        """
        Get the index of the character.

        :returns: The index of the character.
        :rtype: Any
        """
        return self._index()

    def _cnname(self):
        """
        Get the Chinese name of the character.

        This method should be implemented by the subclasses.

        :returns: The Chinese name of the character.
        :rtype: str
        """
        raise NotImplementedError  # pragma: no cover

    def _cnnames(self):
        """
        Get additional Chinese names of the character.

        This method should be implemented by the subclasses.

        :returns: The additional Chinese names of the character.
        :rtype: List[str]
        """
        cnname = self._cnname()
        return [cnname] if cnname else []

    @property
    def cnname(self):
        """
        Get the Chinese name of the character.

        :returns: The Chinese name of the character.
        :rtype: gchar.games.base.ChineseName
        """
        cnname = self._cnname()
        return self.__cnname_class__(cnname) if cnname else None

    @property
    def cnnames(self):
        """
        Get additional Chinese names of the character.

        :returns: The additional Chinese names of the character.
        :rtype: List[gchar.games.base.ChineseName]
        """
        names = [self.__cnname_class__(name) for name in self._cnnames() if name]
        return [name for name in names if name]

    def _jpname(self):
        """
        Get the Japanese name of the character.

        This method should be implemented by the subclasses.

        :returns: The Japanese name of the character.
        :rtype: str
        """
        raise NotImplementedError  # pragma: no cover

    def _jpnames(self):
        """
        Get additional Japanese names of the character.

        This method should be implemented by the subclasses.

        :returns: The additional Japanese names of the character.
        :rtype: List[str]
        """
        jpname = self._jpname()
        return [jpname] if jpname else []

    @property
    def jpname(self):
        """
        Get the Japanese name of the character.

        :returns: The Japanese name of the character.
        :rtype: gchar.games.base.JapaneseName
        """
        jpname = self._jpname()
        return self.__jpname_class__(jpname) if jpname else None

    @property
    def jpnames(self):
        """
        Get additional Japanese names of the character.

        :returns: The additional Japanese names of the character.
        :rtype: List[gchar.games.base.JapaneseName]
        """
        names = [self.__jpname_class__(name) for name in self._jpnames() if name]
        return [name for name in names if name]

    def _enname(self):
        """
        Get the English name of the character.

        This method should be implemented by the subclasses.

        :returns: The English name of the character.
        :rtype: str
        """
        raise NotImplementedError  # pragma: no cover

    def _ennames(self):
        """
        Get additional English names of the character.

        This method should be implemented by the subclasses.

        :returns: The additional English names of the character.
        :rtype: List[str]
        """
        enname = self._enname()
        return [enname] if enname else []

    @property
    def enname(self):
        """
        Get the English name of the character.

        :returns: The English name of the character.
        :rtype: gchar.games.base.EnglishName
        """
        enname = self._enname()
        return self.__enname_class__(enname) if enname else None

    @property
    def ennames(self):
        """
        Get additional English names of the character.

        :returns: The additional English names of the character.
        :rtype: List[gchar.games.base.EnglishName]
        """
        names = [self.__enname_class__(name) for name in self._ennames() if name]
        return [name for name in names if name]

    def _custom_alias_names(self):
        """
        Get custom alias names of the character.

        This method should be implemented by the subclasses.

        :returns: The custom alias names of the character.
        :rtype: List[str]
        """
        return []

    def _generic_alias_names(self):
        """
        Get generic alias names of the character.

        This method should be implemented by the subclasses.

        :returns: The generic alias names of the character.
        :rtype: List[str]
        """
        return sorted(self._get_alias_index().get(self.index, None) or [])

    @property
    def alias_names(self):
        """
        Get all alias names of the character.

        :returns: The alias names of the character.
        :rtype: List[Union[gchar.games.base.ChineseName, gchar.games.base.GenericAliasName]]
        """
        return [
            *(self.__alias_name_class__(name) for name in self._custom_alias_names()),
            *(GenericAliasName(name) for name in self._generic_alias_names()),
        ]

    def _names(self) -> List[_BaseName]:
        """
        Get all names of the character.

        This method should be implemented by the subclasses.

        :returns: The names of the character.
        :rtype: List[_gchar.games.base.BaseName]
        """
        return [*self.cnnames, *self.ennames, *self.jpnames]

    @property
    def names(self) -> List[str]:
        """
        Get all names of the character as strings.

        :returns: The names of the character.
        :rtype: List[str]
        """
        return sorted(set(map(str, self._names())))

    def _gender(self):
        """
        Get the gender of the character.

        This method should be implemented by the subclasses.

        :returns: The gender of the character.
        :rtype: Union[str, Gender]
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def gender(self) -> Gender:
        """
        Get the gender of the character.

        :returns: The gender of the character.
        :rtype: Gender
        """
        return Gender.loads(self._gender())

    def _is_extra(self) -> bool:
        """
        Check if the character is an extra character.

        This method should be implemented by the subclasses.

        :returns: True if the character is an extra character, False otherwise.
        :rtype: bool
        """
        return False

    @property
    def is_extra(self) -> bool:
        """
        Check if the character is an extra character.

        :returns: True if the character is an extra character, False otherwise.
        :rtype: bool
        """
        return bool(self._is_extra())

    def _skins(self) -> List[Tuple[str, str]]:
        """
        Get the skins of the character.

        This method should be implemented by the subclasses.

        :returns: The skins of the character as a list of tuples containing the skin name and URL.
        :rtype: List[Tuple[str, str]]
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def skins(self) -> List[Skin]:
        """
        Get the skins of the character.

        :returns: The skins of the character as a list of Skin objects.
        :rtype: List[Skin]
        """
        return [Skin(self.__game_name__, self._index(), name, url) for name, url in self._skins()]

    def _release_time(self):
        """
        Get the release time of the character.

        This method should be implemented by the subclasses.

        :returns: The release time of the character as a UNIX timestamp, or None if unknown.
        :rtype: Optional[float]
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def release_time(self) -> Optional[float]:
        """
        Get the release time of the character.

        :returns: The release time of the character as a UNIX timestamp, or None if unknown.
        :rtype: Optional[float]
        """
        return self._release_time()

    def _order(self):
        """
        Get the order value for sorting characters.

        This method should be implemented by the subclasses.

        :returns: The order value for sorting characters.
        :rtype: Any
        """
        return ()

    def _key(self):
        """
        Get the key value for comparing characters.

        :returns: The key value for comparing characters.
        :rtype: Tuple[Any, Any, Any]
        """
        return self._order(), self._index(), (1 if self._is_extra() else 0)

    def __eq__(self, other) -> bool:
        """
        Compare the character with another character or a name.

        :param other: The character or name to compare.
        :type other: Union[gchar.games.base.Character, str]
        :returns: True if the character is equal to the other character or name, False otherwise.
        :rtype: bool
        """
        if type(other) == type(self):
            return self.index == other.index
        else:
            if self.index is not None and self.index == other:
                return True
            for name in chain(self._names(), self.alias_names):
                if name == other:
                    return True

            return False

    def __ne__(self, other):
        """
        Compare the character with another character or a name for inequality.

        :param other: The character or name to compare.
        :type other: Union[gchar.games.base.Character, str]
        :returns: True if the character is not equal to the other character or name, False otherwise.
        :rtype: bool
        """
        return not self.__eq__(other)

    @classmethod
    @lru_cache()
    def _get_index(cls) -> List[Dict]:
        """
        Get the index of all characters.

        :returns: The index of all characters.
        :rtype: List[Dict]
        """
        with open(hf_hub_download(
                repo_id=cls.__repository__,
                filename=f'{cls.__game_name__}/index.json',
                repo_type='dataset',
        ), 'r', encoding='utf-8') as f:
            return json.load(f)['data']

    @classmethod
    @lru_cache()
    def _get_alias_index(cls):
        """
        Get the index of alias names for all characters.

        :returns: The index of alias names for all characters.
        :rtype: Dict
        """
        if hf_fs.exists(f'datasets/{cls.__repository__}/{cls.__game_name__}/index_alias.yaml'):
            with open(hf_hub_download(
                    repo_id=cls.__repository__,
                    filename=f'{cls.__game_name__}/index_alias.yaml',
                    repo_type='dataset',
            ), 'r', encoding='utf-8') as f:
                data = yaml.load(f, Loader)

            return {
                item['id']: item['alias']
                for item in data['alias']
            }

        else:
            return {}

    @classmethod
    @lru_cache()
    def _simple_all(cls, contains_extra: bool = True) -> List:
        """
        Get all characters.

        :param contains_extra: Whether to include extra characters in the result.
        :type contains_extra: bool
        :returns: All characters.
        :rtype: List[gchar.games.base.Character]
        """
        all_chs = [cls(data) for data in cls._get_index()]
        chs = [ch for ch in all_chs if contains_extra or not ch.is_extra]
        return chs

    @classmethod
    def all(cls, contains_extra: bool = True, sorted: bool = True) -> List:
        """
        Get all characters.

        :param contains_extra: Whether to include extra characters in the result.
        :type contains_extra: bool
        :param sorted: Whether to sort the characters by order.
        :type sorted: bool
        :returns: All characters.
        :rtype: List[gchar.games.base.Character]
        """
        chs = cls._simple_all(contains_extra)
        if sorted:
            return builtins.sorted(chs)
        else:
            return chs

    @classmethod
    def get(cls, name, **kwargs):
        """
        Get a character by name.

        :param name: The name of the character.
        :type name: str
        :returns: The character with the given name, or None if not found.
        :rtype: Optional[gchar.games.base.Character]
        """
        for item in cls._simple_all(**kwargs):
            if item == name:
                return item

        return None


def _yield_all_characters(ch: Union[Character, list, tuple, Type[Character]], **kwargs) -> Iterator[Character]:
    """
    Helper function to recursively yield all characters.

    :param ch: The character or list of characters.
    :type ch: Union[gchar.games.base.Character, list, tuple, Type[gchar.games.base.Character]]
    :returns: An iterator over all characters.
    :rtype: Iterator[gchar.games.base.Character]
    """
    if isinstance(ch, Character):
        yield ch
    elif isinstance(ch, type) and issubclass(ch, Character):
        yield from ch.all(**kwargs)
    elif isinstance(ch, (list, tuple)):
        for item in ch:
            yield from _yield_all_characters(item, **kwargs)


def list_all_characters(*chs: Union[Character, list, tuple, Type[Character]], **kwargs) -> List[Character]:
    """
    List all characters.

    :param chs: The characters or lists of characters.
    :type chs: Union[gchar.games.base.Character, list, tuple, Type[gchar.games.base.Character]]
    :param contains_extra: Whether to include extra characters in the result.
    :type contains_extra: bool
    :returns: All characters.
    :rtype: List[gchar.games.base.Character]
    """
    return list(_yield_all_characters(chs, **kwargs))



from .cache import optional_lru_cache
from .cli import print_version, GLOBAL_CONTEXT_SETTINGS
from .compare import Comparable
from .download import download_file
from .session import get_requests_session, sget, srequest



from functools import lru_cache
from itertools import islice
from typing import Iterator, Tuple, Optional, List, Type, Dict

from thefuzz import fuzz

from ..arknights import Character as ArknightsCharacter
from ..azurlane import Character as AzurLaneCharacter
from ..base import Character
from ..bluearchive import Character as BlueArchiveCharacter
from ..fgo import Character as FateGrandOrderCharacter
from ..genshin import Character as GenshinImpactCharacter
from ..girlsfrontline import Character as GirlsFrontLineCharacter
from ..neuralcloud import Character as NeuralCloudCharacter
from ..nikke import Character as NikkeCharacter
from ..pathtonowhere import Character as PathToNowhereCharacter
from ..starrail import Character as StarRailCharacter
from ...utils import optional_lru_cache

GAME_CHARS: Dict[str, Type[Character]] = {}


def register_game(cls: Type[Character]):
    """
    Register a game and its corresponding character class.

    :param cls: The character class.
    :type cls: Type[Character]
    :raises KeyError: If the game is already registered.
    """
    if cls.__game_name__ in GAME_CHARS:
        raise KeyError(f'Game {cls.__game_name__!r} already exist.')
    else:
        GAME_CHARS[cls.__game_name__] = cls


def list_available_game_names() -> List[str]:
    """
    Get a list of available game names.

    :return: The list of game names.
    :rtype: List[str]

    Examples::
        >>> from gchar.games import list_available_game_names
        >>>
        >>> list_available_game_names()
        ['arknights', 'azurlane', 'bluearchive', 'fgo', 'genshin', 'girlsfrontline', 'neuralcloud', 'nikke', 'pathtonowhere', 'starrail']
    """
    return sorted(GAME_CHARS.keys())


def get_character_class(game_name: str) -> Type[Character]:
    """
    Get the character class for a given game name.

    :param game_name: The name of the game.
    :type game_name: str
    :return: The character class.
    :rtype: Type[Character]
    :raises KeyError: If the game is not found.
    """
    if game_name in GAME_CHARS:
        return GAME_CHARS[game_name]
    else:
        raise KeyError(f'Game {game_name!r} not found.')


register_game(ArknightsCharacter)
register_game(FateGrandOrderCharacter)
register_game(AzurLaneCharacter)
register_game(GenshinImpactCharacter)
register_game(GirlsFrontLineCharacter)
register_game(NeuralCloudCharacter)
register_game(BlueArchiveCharacter)
register_game(PathToNowhereCharacter)
register_game(NikkeCharacter)
register_game(StarRailCharacter)


@optional_lru_cache()
def _all_characters(**kwargs) -> List[Character]:
    """
    Get a list of all characters from all registered games.

    :param kwargs: Additional arguments for character retrieval.
    :return: The list of characters.
    :rtype: List[Character]
    """
    from ...resources.pixiv import get_pixiv_posts

    chs: List[Tuple[Character, int, bool, int]] = []
    cnt = 0
    for _, _ch_set in GAME_CHARS.items():
        for ch in _ch_set.all(**kwargs, sorted=False):
            counts = get_pixiv_posts(ch)
            if counts:
                all_count, _ = counts
            else:
                all_count = 0
            chs.append((ch, all_count, ch.is_extra, cnt))
            cnt += 1

    chs = sorted(chs, key=lambda x: (0 if not x[2] else 1, -x[1], x[3]))
    return [ch for ch, _, _, _ in chs]


# NAME SIMILAR: 101-200
# ALIAS SIMILAR: 1-100
_ALIAS_MATCH = 300
_NAME_MATCH = 400


def _name_trim(name: str) -> str:
    """
    Trim a character name by removing spaces, dashes, underscores, and special characters.

    :param name: The character name.
    :type name: str
    :return: The trimmed character name.
    :rtype: str
    """
    return name.replace(' ', '').replace('-', '').replace('_', '').replace(chr(160), '').replace('\t', '')


def _yield_characters(name: str, allow_fuzzy: bool = False, fuzzy_threshold: int = 80, **kwargs) \
        -> Iterator[Tuple[Character, int]]:
    """
    Yield characters that match the given name.

    :param name: The character name.
    :type name: str
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param fuzzy_threshold: The threshold for fuzzy matching.
    :type fuzzy_threshold: int
    :param kwargs: Additional arguments for character retrieval.
    :return: The iterator of matching characters and their match scores.
    :rtype: Iterator[Tuple[Character, int]]
    """
    all_chs = _all_characters(**kwargs)
    for ch in all_chs:
        if any([_name == name for _name in ch._names()]):
            yield ch, _NAME_MATCH
    for ch in all_chs:
        if any([_name == name for _name in ch.alias_names]):
            yield ch, _ALIAS_MATCH

    if allow_fuzzy:
        @lru_cache()
        def _get_formatted_name(cls) -> str:
            return str(cls(name))

        items: List[Tuple[Character, int, int]] = []
        for i, ch in enumerate(all_chs):
            name_fuzzs = [
                fuzz.WRatio(
                    _name_trim(_get_formatted_name(type(_name))),
                    _name_trim(str(_name)),
                    force_ascii=False
                )
                for _name in ch._names()
            ]
            if name_fuzzs and max(name_fuzzs) >= fuzzy_threshold:
                items.append((ch, max(name_fuzzs) + 100, i))
        items = sorted(items, key=lambda x: (-x[1], x[2]))
        for ch, fuzzy, _ in items:
            yield ch, fuzzy

        items: List[Tuple[Character, int, int]] = []
        for i, ch in enumerate(all_chs):
            alias_fuzzs = [
                fuzz.WRatio(
                    _name_trim(_get_formatted_name(type(_name))),
                    _name_trim(str(_name)),
                    force_ascii=False
                )
                for _name in ch.alias_names
            ]
            if alias_fuzzs and max(alias_fuzzs) >= fuzzy_threshold:
                items.append((ch, max(alias_fuzzs), i))
        items = sorted(items, key=lambda x: (-x[1], x[2]))
        for ch, fuzzy, _ in items:
            yield ch, fuzzy


def list_character(name: str, limit: Optional[int] = None, allow_fuzzy: bool = False,
                   fuzzy_threshold: int = 80, **kwargs) -> List[Character]:
    """
    List characters that match the given name.

    :param name: The character name.
    :type name: str
    :param limit: The maximum number of characters to return.
    :type limit: Optional[int]
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param fuzzy_threshold: The threshold for fuzzy matching.
    :type fuzzy_threshold: int
    :param kwargs: Additional arguments for character retrieval.
    :return: The list of matching characters.
    :rtype: List[Character]
    """
    iterator = _yield_characters(name, allow_fuzzy, fuzzy_threshold, **kwargs)
    if limit is not None:
        iterator = islice(iterator, limit)

    ordered = [ch for ch, _ in iterator]
    assert limit is None or len(ordered) <= limit
    return ordered


def get_character(name: str, allow_fuzzy: bool = False, fuzzy_threshold: int = 80,
                  contains_extra: bool = True, **kwargs) -> Optional[Character]:
    """
    Get a character that matches the given name.

    :param name: The character name.
    :type name: str
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param fuzzy_threshold: The threshold for fuzzy matching.
    :type fuzzy_threshold: int
    :param contains_extra: Whether to include extra characters.
    :type contains_extra: bool
    :param kwargs: Additional arguments for character retrieval.
    :return: The matching character, or None if not found.
    :rtype: Optional[Character]

    Examples::
        >>> from gchar.games import get_character
        >>>
        >>> get_character('amiya')
        <Character R001 - 阿米娅/amiya/アーミヤ, female, 5*****>
        >>> get_character('surtr')
        <Character R111 - 史尔特尔/surtr/スルト, female, 6******>
        >>> get_character('dusk')
        <Character NM02 - 夕/dusk/シー, female, 6******>
        >>> get_character('yae_miko')
        <Character 八重神子/yae_miko/八重神子/やえみこ, female, 5*****, weapon: Weapon.CATALYST, element: Element.ELECTRO>
    """
    _items = list_character(name, limit=1, allow_fuzzy=allow_fuzzy, fuzzy_threshold=fuzzy_threshold,
                            contains_extra=contains_extra, **kwargs)
    if _items:
        return _items[0]
    else:
        return None



import re
from functools import lru_cache
from typing import List, Optional, Tuple

from .name import ChineseName, JapaneseName, EnglishName, ChineseAliasName
from .property import Rarity, Clazz
from ..base import Character as _BaseCharacter


class Character(_BaseCharacter):
    __game_name__ = 'fgo'
    __official_name__ = 'Fate/Grand Order'
    __cnname_class__ = ChineseName
    __jpname_class__ = JapaneseName
    __enname_class__ = EnglishName
    __alias_name_class__ = ChineseAliasName

    def __init__(self, raw_data: dict):
        self.__raw_data = raw_data

    def _index(self) -> int:
        return self.__raw_data['id']

    def _gender(self):
        return self.__raw_data['gender']

    @property
    def rarity(self) -> Rarity:
        return Rarity.loads(int(self.__raw_data['rarity']))

    @property
    def accessible(self) -> bool:
        return self.__raw_data['accessible']

    @property
    def clazz(self) -> Clazz:
        try:
            return Clazz.loads(self.__raw_data['class'])
        except (KeyError, TypeError, ValueError):
            return self.__raw_data['class']

    def _cnname(self):
        return self._cnnames()[0]

    def _cnnames(self):
        return self.__raw_data['cnnames']

    def _jpname(self):
        return self._jpnames()[0]

    def _jpnames(self):
        original_jpnames = list(self.__raw_data['jpnames'])
        external_names = []
        for name in original_jpnames:
            matching = re.fullmatch(r'^(?P<name>[\s\S]+?)〔(?P<suffix>[\s\S]+?)〕$', name)
            if matching:
                suffixes = re.findall(r'\b\w+\b', matching.group('suffix'))
                external_names.append('・'.join([matching.group('name'), *suffixes]))

                name_segments = re.findall(r'\b\w+\b', matching.group('name'))
                if len(name_segments) > 1:
                    external_names.append('・'.join([name_segments[0], *suffixes]))

        original_jpnames.extend(external_names)
        return original_jpnames

    def _enname(self):
        return self._ennames()[0]

    def _ennames(self):
        return self.__raw_data['ennames']

    def _custom_alias_names(self):
        return self.__raw_data['alias']

    def _is_extra(self) -> bool:
        for ch in self.__raw_data['similar']:
            id_ = ch['id']
            the_ch = self._get_by_id(id_)
            if id_ < self.__raw_data['id'] and the_ch is not None and \
                    any([name == the_ch for name in self.names]):
                return True

        return False

    def _release_time(self):
        return None

    def __repr__(self):
        return f'<{type(self).__name__} {self.index} - {"/".join(map(str, self._names()))}, ' \
               f'{self.gender.name.lower()}, {self.rarity}{"*" * self.rarity}, class: {self.clazz}>'

    def _skins(self) -> List[Tuple[str, str]]:
        return [(item['name'], item['url']) for item in self.__raw_data['skins']]

    @classmethod
    @lru_cache()
    def _make_index_by_id(cls):
        all_chs = cls._simple_all(contains_extra=True)
        return {ch.index: ch for ch in all_chs}

    @classmethod
    def _get_by_id(cls, index: int) -> Optional['Character']:
        return cls._make_index_by_id().get(index, None)



from enum import unique, IntEnum, Enum


@unique
class Group(Enum):
    """
    An enumeration representing the groups in the Azur Lane game.
    """

    USS = 0x1
    HMS = 0x2
    IJN = 0x3
    KMS = 0x4
    DE = 0x5
    RN = 0x6
    SN = 0x7
    FFNF = 0x8
    MNF = 0x9

    def __eq__(self, other):
        if isinstance(other, Group):
            return self.value == other.value
        else:
            try:
                return self == Group.loads(other)
            except (TypeError, ValueError):
                return False

    @classmethod
    def loads(cls, val) -> 'Group':
        """
        Load a Group value from a string or an int.

        :param val: The value to load.
        :type val: Union[str, int]
        :return: The loaded Group value.
        :rtype: Group
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            if val == '白鹰':
                return cls.USS
            elif val == '皇家':
                return cls.HMS
            elif val == '重樱':
                return cls.IJN
            elif val == '铁血':
                return cls.KMS
            elif val == '东煌':
                return cls.DE
            elif val == '撒丁帝国':
                return cls.RN
            elif val == '北方联合':
                return cls.SN
            elif val == '自由鸢尾':
                return cls.FFNF
            elif val == '维希教廷':
                return cls.MNF
            else:
                raise ValueError(f'Invalid group value - {val!r}.')
        else:
            raise TypeError(f'Invalid group type - {val!r}.')


@unique
class BasicRarity(IntEnum):
    """
    An enumeration representing the basic rarity levels in the Azur Lane game.
    """

    COMMON = 0x1
    RARE = 0x2
    ELITE = 0x3
    ULTRA = 0x4
    EPIC = 0x5

    @property
    def label(self) -> str:
        """
        Get the label corresponding to the rarity level.

        :return: The label of the rarity level.
        :rtype: str
        """
        if self == self.COMMON:
            return '普通'
        elif self == self.RARE:
            return '稀有'
        elif self == self.ELITE:
            return '精锐'
        elif self == self.ULTRA:
            return '超稀有'
        elif self == self.EPIC:
            return '海上传奇'
        else:
            raise ValueError(f'Unknown basic level - {self!r}.')  # pragma: no cover

    def __eq__(self, other):
        if isinstance(other, BasicRarity):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def loads(cls, val) -> 'BasicRarity':
        """
        Load a BasicRarity value from a string or an int.

        :param val: The value to load.
        :type val: Union[str, int]
        :return: The loaded BasicRarity value.
        :rtype: BasicRarity
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item

            raise ValueError(f'Invalid level value - {val!r}.')
        elif isinstance(val, str):
            if val == '普通':
                return cls.COMMON
            elif val == '稀有':
                return cls.RARE
            elif val == '精锐':
                return cls.ELITE
            elif val == '超稀有':
                return cls.ULTRA
            elif val == '海上传奇':
                return cls.EPIC
            else:
                raise ValueError(f'Invalid level string - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')


@unique
class ResearchRarity(IntEnum):
    """
    An enumeration representing the research rarity levels in the Azur Lane game.
    """

    TOP = 0x4
    DECISIVE = 0x5

    @property
    def label(self) -> str:
        """
        Get the label corresponding to the research rarity level.

        :return: The label of the research rarity level.
        :rtype: str
        """
        if self == self.TOP:
            return '最高方案'
        elif self == self.DECISIVE:
            return '决战方案'
        else:
            raise ValueError(f'Unknown research level - {self!r}.')  # pragma: no cover

    def __eq__(self, other):
        if isinstance(other, ResearchRarity):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def loads(cls, val) -> 'ResearchRarity':
        """
        Load a ResearchRarity value from a string or an int.

        :param val: The value to load.
        :type val: Union[str, int]
        :return: The loaded ResearchRarity value.
        :rtype: ResearchRarity
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item

            raise ValueError(f'Invalid level value - {val!r}.')
        elif isinstance(val, str):
            if val == '最高方案':
                return cls.TOP
            elif val == '决战方案':
                return cls.DECISIVE
            else:
                raise ValueError(f'Invalid level string - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')



"""
Overview:
    Character basic modeling, including the character model itself, the character name,
    the character gender, and the character skins.
"""
from .character import Character, list_all_characters
from .name import ChineseName, EnglishName, JapaneseName, TextName, SegmentName
from .property import Gender
from .skin import Skin



from typing import List, Tuple

import math

from .name import EnglishName, JapaneseName, ChineseName, ChineseAliasName
from .property import Rarity, Clazz
from ..base import Character as _BaseCharacter


class Character(_BaseCharacter):
    __game_name__ = 'girlsfrontline'
    __official_name__ = 'Girls\' Frontline'
    __enname_class__ = EnglishName
    __cnname_class__ = ChineseName
    __jpname_class__ = JapaneseName
    __alias_name_class__ = ChineseAliasName

    def __init__(self, raw_data: dict):
        self.__raw_data = raw_data

    def _index(self) -> int:
        return self.__raw_data['id']

    def _cnname(self):
        return self.__raw_data['cnname']

    def _cnnames(self):
        return self.__raw_data['cnnames']

    def _enname(self):
        return self.__raw_data['enname']

    def _jpname(self):
        return self.__raw_data['jpname']

    def _custom_alias_names(self):
        return self.__raw_data['alias']

    def _gender(self):
        return 'female'

    @property
    def rarity(self) -> Rarity:
        return Rarity.loads(self.__raw_data['rarity'])

    @property
    def clazz(self) -> Clazz:
        return Clazz.loads(self.__raw_data['class'])

    def _skins(self) -> List[Tuple[str, str]]:
        return [(item['desc'], item['url']) for item in self.__raw_data['skins']]

    def _release_time(self):
        release_info = self.__raw_data['release']
        return release_info['time']

    def _order(self):
        _release_time = self._release_time()
        return _release_time if _release_time is not None else math.inf

    def __repr__(self):
        if isinstance(self.rarity.value, int):
            _rarity_repr = f'{self.rarity.value}{"*" * self.rarity.value}'
        else:
            _rarity_repr = f'{self.rarity.name}'

        return f'<{type(self).__name__} {self.index} - {"/".join(map(str, self._names()))}, ' \
               f'{_rarity_repr}, clazz: {self.clazz}>'



from enum import unique, Enum


@unique
class Gender(Enum):
    """
    Enumeration class representing the gender of a character.

    The class provides three gender options: OTHER, MALE, and FEMALE.
    It supports equality comparison with other Gender objects and string values.
    The class also provides a method `loads` for parsing and converting string values into Gender objects.

    Usage:
        * Use the Gender enum to represent the gender of a character.

    Supported string values for each gender:
        * OTHER: "OTHER"
        * MALE: "MALE", "男", "男性", "boy", "man", "male"
        * FEMALE: "FEMALE", "女", "女性", "girl", "woman", "female"
    """
    OTHER = 0x0
    MALE = 0x1
    FEMALE = 0x2

    def __eq__(self, other):
        """
        Equality comparison with other Gender objects or string values.

        :param other: The other object to compare.
        :type other: object
        :returns: True if the objects are equal, False otherwise.
        :rtype: bool
        """
        if isinstance(other, Gender):
            return other.value == self.value
        else:
            try:
                other = Gender.loads(other)
            except (TypeError, ValueError):
                return False
            else:
                return other == self

    @classmethod
    def loads(cls, val) -> 'Gender':
        """
        Parse and convert a string value into a Gender object.

        The method accepts string values representing genders and converts them into the corresponding Gender enum object.
        If the input value is already a Gender object, it is returned as is.

        :param val: The value to parse and convert.
        :type val: Union[str, Gender]
        :returns: The Gender object corresponding to the input value.
        :rtype: Gender
        :raises TypeError: If the input value is of an invalid type.
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            nval = val.lower().strip()
            if nval in {'男', '男性', 'boy', 'man', 'male'}:
                return cls.MALE
            elif nval in {'女', '女性', 'girl', 'woman', 'female'}:
                return cls.FEMALE
            else:
                return cls.OTHER
        else:
            raise TypeError(f'Invalid gender type - {val!r}.')



from typing import List, Optional, Tuple

from .name import EnglishName, JapaneseName, ChineseName
from .property import Rarity, Weapon, Element
from ..base import Character as _BaseCharacter


class Character(_BaseCharacter):
    __game_name__ = 'genshin'
    __official_name__ = 'Genshin Impact'
    __cnname_class__ = ChineseName
    __enname_class__ = EnglishName
    __jpname_class__ = JapaneseName

    def __init__(self, raw_data: dict):
        self.__raw_data = raw_data

    def _index(self) -> str:
        return str(self.enname)

    def _gender(self):
        return self.__raw_data['gender']

    def _cnname(self):
        return self.__raw_data['cnname']

    def _enname(self):
        return self.__raw_data['enname']

    def _jpname(self):
        _jpnames = self._jpnames()
        if _jpnames:
            return _jpnames[0]
        else:
            return None  # pragma: no cover

    def _jpnames(self):
        return self.__raw_data['jpnames']

    @property
    def rarity(self) -> Rarity:
        return Rarity.loads(self.__raw_data['rarity'])

    @property
    def weapon(self) -> Weapon:
        return Weapon.loads(self.__raw_data['weapon'])

    @property
    def element(self) -> Optional[Element]:
        element = self.__raw_data['element']
        if element:
            return Element.loads(element)
        else:
            return None

    def _skins(self) -> List[Tuple[str, str]]:
        return [(item['name'], item['url']) for item in self.__raw_data['skins']]

    def _release_time(self):
        release_info = self.__raw_data['release']
        return release_info['time']

    def _order(self):
        return self._release_time()

    def __repr__(self):
        return f'<{type(self).__name__} {"/".join(map(str, self._names()))}, ' \
               f'{self.gender.name.lower()}, {self.rarity}{"*" * self.rarity}, ' \
               f'weapon: {self.weapon}, element: {self.element}>'



from typing import List, Tuple

from .name import ChineseName, EnglishName
from .property import Rarity
from ..base.character import Character as _BaseCharacter


class Character(_BaseCharacter):
    __game_name__ = 'pathtonowhere'
    __official_name__ = 'Path To Nowhere'
    __cnname_class__ = ChineseName
    __enname_class__ = EnglishName

    def __init__(self, raw_data):
        self.__raw_data = raw_data

    def _index(self):
        return self.__raw_data['id']

    def _cnname(self):
        return self.__raw_data['cnname']

    def _jpname(self):
        return []

    def _enname(self):
        return self.__raw_data['enname']

    def _gender(self):
        return self.__raw_data['gender']

    def _skins(self) -> List[Tuple[str, str]]:
        return [(skin['name'], skin['url']) for skin in self.__raw_data['skins']]

    def _release_time(self):
        return None

    @property
    def rarity(self) -> Rarity:
        return Rarity.loads(self.__raw_data['rarity'])

    @property
    def job(self):
        return self.__raw_data['job']

    @property
    def group(self):
        return self.__raw_data['group']

    def __repr__(self):
        return f'<{type(self).__name__} {self._index()} - {"/".join(map(str, self._names()))}, ' \
               f'{self.rarity.name}({self.rarity.number}{"*" * self.rarity.number}), ' \
               f'job: {self.job}, group: {self.group}>'



from enum import IntEnum, unique, Enum


@unique
class Rarity(IntEnum):
    """
    Enum representing the rarity of FGO characters.

    :cvar ZERO: Rarity 0
    :cvar ONE: Rarity 1
    :cvar TWO: Rarity 2
    :cvar THREE: Rarity 3
    :cvar FOUR: Rarity 4
    :cvar FIVE: Rarity 5
    """

    ZERO = 0x0
    ONE = 0x1
    TWO = 0x2
    THREE = 0x3
    FOUR = 0x4
    FIVE = 0x5

    @classmethod
    def loads(cls, val) -> 'Rarity':
        """
        Load a rarity from a value.

        :param val: The value to load the rarity from.
        :type val: Union[int, :class:`Rarity`]
        :returns: The loaded rarity.
        :rtype: :class:`Rarity`
        :raises ValueError: If the value is invalid.
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item.value

            raise ValueError(f'Invalid level value - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')


@unique
class Clazz(Enum):
    """
    Enum representing the class of FGO characters.

    :cvar SABER: Saber class
    :cvar LANCER: Lancer class
    :cvar ARCHER: Archer class
    :cvar RIDER: Rider class
    :cvar CASTER: Caster class
    :cvar ASSASSIN: Assassin class
    :cvar BERSERKER: Berserker class
    :cvar AVENGER: Avenger class
    :cvar RULER: Ruler class
    :cvar MOONCANCER: MoonCancer class
    :cvar SHIELDER: Shielder class
    :cvar ALTEREGO: AlterEgo class
    :cvar PRETENDER: Pretender class
    :cvar FOREIGNER: Foreigner class
    """

    SABER = 0x1
    LANCER = 0x2
    ARCHER = 0x3
    RIDER = 0x4
    CASTER = 0x5
    ASSASSIN = 0x6
    BERSERKER = 0x7
    AVENGER = 0x8
    RULER = 0x9
    MOONCANCER = 0xa
    SHIELDER = 0xb
    ALTEREGO = 0xc
    PRETENDER = 0xd
    FOREIGNER = 0xe

    def __eq__(self, other):
        """
        Compare the class with another class.

        :param other: The other class to compare.
        :type other: Union[:class:`Clazz`, str]
        :returns: True if the classes are equal, False otherwise.
        :rtype: bool
        """
        if isinstance(other, Clazz):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False

    @classmethod
    def loads(cls, val) -> 'Clazz':
        """
        Load a class from a value.

        :param val: The value to load the class from.
        :type val: Union[str, :class:`Clazz`]
        :returns: The loaded class.
        :rtype: :class:`Clazz`
        :raises ValueError: If the value is invalid.
        """
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            try:
                return cls.__members__[val.upper().strip()]
            except KeyError:
                raise ValueError(f'Invalid class value - {val!r}.')
        else:
            raise TypeError(f'Invalid class type - {val!r}.')



from typing import List, Optional, Union, Tuple

from .name import EnglishName, JapaneseName, ChineseName, ChineseAliasName
from .property import BasicRarity, ResearchRarity, Group
from ..base import Character as _BaseCharacter


class Character(_BaseCharacter):
    """
    A class representing characters in the Azur Lane game.
    Inherits from _BaseCharacter.
    """

    __game_name__ = 'azurlane'
    __official_name__ = 'Azur Lane'
    __cnname_class__ = ChineseName
    __enname_class__ = EnglishName
    __jpname_class__ = JapaneseName
    __alias_name_class__ = ChineseAliasName

    def __init__(self, raw_data: dict):
        """
        Initialize a Character object.

        :param raw_data: The raw data dictionary representing the character.
        :type raw_data: dict
        """
        self.__raw_data = raw_data

    def _index(self) -> str:
        """
        Get the index of the character.

        :return: The index of the character.
        :rtype: str
        """
        return self.__raw_data['id']

    def _cnname(self):
        """
        Get the Chinese name of the character.

        :return: The Chinese name of the character.
        :rtype: str
        """
        return self.__raw_data['cnname']['short']

    def _enname(self):
        """
        Get the English name of the character.

        :return: The English name of the character.
        :rtype: str
        """
        return self.__raw_data['enname']['short']

    def _jpname(self):
        """
        Get the primary Japanese name of the character.

        :return: The primary Japanese name of the character.
        :rtype: str
        """
        if self.__raw_data['jpnames']:
            return self.__raw_data['jpnames'][0]
        else:
            return None

    def _jpnames(self):
        """
        Get the list of all Japanese names of the character.

        :return: The list of all Japanese names of the character.
        :rtype: List[str]
        """
        return self.__raw_data['jpnames']

    def _custom_alias_names(self):
        """
        Get the list of custom alias names of the character.

        :return: The list of custom alias names of the character.
        :rtype: List[str]
        """
        return self.__raw_data['alias']

    def _gender(self):
        """
        Get the gender of the character.

        :return: The gender of the character.
        :rtype: str
        """
        return 'female'

    @property
    def rarity(self) -> Optional[Union[BasicRarity, ResearchRarity]]:
        """
        Get the rarity of the character.

        :return: The rarity of the character.
        :rtype: Optional[Union[BasicRarity, ResearchRarity]]
        """
        val = self.__raw_data['rarity']
        try:
            return BasicRarity.loads(val)
        except (ValueError, TypeError):
            return ResearchRarity.loads(val)

    @property
    def group(self) -> Union[Group, str]:
        """
        Get the group of the character.

        :return: The group of the character.
        :rtype: Union[Group, str]
        """
        try:
            return Group.loads(self.__raw_data['group'])
        except (TypeError, ValueError):
            return self.__raw_data['group']

    @property
    def is_meta(self) -> bool:
        """
        Check if the character is a meta character.

        :return: True if the character is a meta character, False otherwise.
        :rtype: bool
        """
        return self.__raw_data['is_meta']

    @property
    def is_refit(self) -> bool:
        """
        Check if the character is a refit character.

        :return: True if the character is a refit character, False otherwise.
        :rtype: bool
        """
        return self.__raw_data['is_refit']

    @property
    def is_mu(self) -> bool:
        """
        Check if the character is a Mu character.

        :return: True if the character is a Mu character, False otherwise.
        :rtype: bool
        """
        return self.__raw_data['is_mu']

    @property
    def is_chibi(self) -> bool:
        """
        Check if the character has a chibi form.

        :return: True if the character has a chibi form, False otherwise.
        :rtype: bool
        """
        return self.__raw_data['is_chibi']

    def _is_extra(self) -> bool:
        """
        Check if the character is considered an "extra" character.

        :return: True if the character is considered an "extra" character, False otherwise.
        :rtype: bool
        """
        return self.is_meta or self.is_refit or self.is_mu or self.is_chibi

    def _skins(self) -> List[Tuple[str, str]]:
        """
        Get the list of skins of the character.

        :return: The list of skins of the character, each represented as a tuple of (name, url).
        :rtype: List[Tuple[str, str]]
        """
        return [(item['name'], item['url']) for item in self.__raw_data['skins']]

    def _release_time(self):
        """
        Get the release time of the character.

        :return: The release time of the character.
        :rtype: Optional[float]
        """
        release_info = self.__raw_data['release']
        return release_info['time']

    def _order(self):
        """
        Get the ordering key of the character.

        :return: The ordering key of the character.
        :rtype: float
        """
        return self._release_time()

    def __repr__(self):
        """
        Get the string representation of the character.

        :return: The string representation of the character.
        :rtype: str
        """
        return f'<{type(self).__name__} {self.index} - {"/".join(map(str, self._names()))}, ' \
               f'{self.rarity.label}({int(self.rarity)}{"*" * self.rarity}), ' \
               f'group: {self.group if isinstance(self.group, Group) else self.group}>'



import json
from functools import lru_cache
from typing import Type, Dict, Tuple, Optional, List, Union

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
import yaml
from huggingface_hub import hf_hub_download, hf_hub_url

from ...games import get_character
from ...games.base import Character
from ...utils import get_requests_session, srequest


@lru_cache()
def _load_pixiv_alias_for_game(cls: Type[Character]) -> Dict[Union[int, str], List[str]]:
    """
    Load the Pixiv aliases for a specific game's character class.

    :param cls: The character class.
    :type cls: Type[Character]
    :returns: A dictionary mapping character IDs or names to a list of Pixiv aliases.
    :rtype: Dict[Union[int, str], List[str]]
    """
    game_name = cls.__game_name__
    resource_url = hf_hub_url(
        cls.__repository__,
        filename=f'{game_name}/pixiv_alias.yaml',
        repo_type='dataset'
    )
    session = get_requests_session()
    if srequest(session, 'HEAD', resource_url, raise_for_status=False).ok:
        with open(hf_hub_download(
                cls.__repository__,
                filename=f'{game_name}/pixiv_alias.yaml',
                repo_type='dataset'
        ), 'r', encoding='utf-8') as f:
            data = yaml.load(f, Loader)

        alias = data['alias']
        return {item['id']: item['alias'] for item in alias}

    else:
        return {}


def _parse_pixiv_names_file(names) -> Dict[str, Tuple[int, float, List[Tuple[str, int]]]]:
    """
    Parse the Pixiv names file and return a dictionary with character information.

    :param names: The names data from the Pixiv names file.
    :type names: Any
    :returns: A dictionary containing character names, counts, pollution ratio, and pollution words.
    :rtype: Dict[str, Tuple[int, float, List[Tuple[str, int]]]]
    """
    all_dict = {item['name']: item for item in names}
    all_counts = {}

    def _get_count(name):
        if name in all_counts:
            return all_counts[name]
        elif name not in all_dict:
            return 0
        else:
            item = all_dict[name]
            cnt = item['count']
            for ex in item['keyword']['excluded']:
                cnt += _get_count(ex)

            all_counts[name] = cnt
            return cnt

    retval = {}
    for item in names:
        name = item['name']
        count = _get_count(name)
        kwinfo = item['keyword']

        p_total = kwinfo['items']
        p_polluted = kwinfo['polluted']
        pollution_ratio = (p_polluted / p_total) if p_polluted else 0.0
        pollition_words = [(word, int(pcnt / p_total * count))
                           for word, pcnt in kwinfo['pollution'].items() if pcnt > 0]

        retval[name] = (count, pollution_ratio, pollition_words)

    return retval


def _load_pixiv_names_for_game(cls: Type[Character]) -> Dict[str, Tuple[int, float, List[Tuple[str, int]]]]:
    """
    Load the Pixiv names for a specific game's character class.

    :param cls: The character class.
    :type cls: Type[Character]
    :returns: A dictionary containing character names, counts, pollution ratio, and pollution words.
    :rtype: Dict[str, Tuple[int, float, List[Tuple[str, int]]]]
    """
    game_name = cls.__game_name__
    with open(hf_hub_download(
            cls.__repository__,
            filename=f'{game_name}/pixiv_names.json',
            repo_type='dataset'
    ), 'r', encoding='utf-8') as f:
        data = json.load(f)
        names = data['names']

    return _parse_pixiv_names_file(names)


@lru_cache()
def _load_pixiv_characters_for_game(cls: Type[Character]) -> Dict[str, Tuple[int, int]]:
    """
    Load the Pixiv character data for a specific game's character class.

    :param cls: The character class.
    :type cls: Type[Character]
    :returns: A dictionary mapping character indices to the number of all illustrations and R18 illustrations.
    :rtype: Dict[str, Tuple[int, int]]
    """
    game_name = cls.__game_name__
    with open(hf_hub_download(
            cls.__repository__,
            filename=f'{game_name}/pixiv_characters.json',
            repo_type='dataset'
    ), 'r', encoding='utf-8') as f:
        chs = json.load(f)['characters']

    return {item['index']: (item["illustrations"]["all"], item["illustrations"]["r18"]) for item in chs}


def get_pixiv_posts(char: Union[str, Character], allow_fuzzy: bool = True, fuzzy_threshold: int = 70, **kwargs) \
        -> Optional[Tuple[int, int]]:
    """
    Get the number of Pixiv posts for a given character.

    :param char: The character name or instance of Character class.
    :type char: Union[str, Character]
    :param allow_fuzzy: Whether to allow fuzzy matching of character names.
    :type allow_fuzzy: bool
    :param fuzzy_threshold: The threshold for fuzzy matching.
    :type fuzzy_threshold: int
    :returns: A tuple containing the number of all illustrations and R18 illustrations, or None if the character is unknown.
    :rtype: Optional[Tuple[int, int]]
    :raises ValueError: If the character is unknown.

    Examples::
        >>> from gchar.resources.pixiv import get_pixiv_posts
        >>>
        >>> get_pixiv_posts('amiya')
        (14130, 1113)
        >>> get_pixiv_posts('surtr')
        (5579, 787)
    """
    original_char = char
    if not isinstance(char, Character):
        char = get_character(char, allow_fuzzy, fuzzy_threshold, **kwargs)
    if not char:
        raise ValueError(f'Unknown character - {original_char!r}.')

    all_data = _load_pixiv_characters_for_game(type(char))
    return all_data.get(char.index, None)
